"""
Python Lexical Analyser

Classes for building NFAs and DFAs
"""

import cython
from .Transitions import TransitionMap

maxint = 2**31-1  # sentinel value

LOWEST_PRIORITY = -maxint


class Machine:
    """A collection of Nodes representing an NFA or DFA."""
    def __init__(self):
        self.states = []  # [Node]
        self.initial_states = {}  # {(name, bol): Node}
        self.next_state_number = 1

    def __del__(self):
        for state in self.states:
            state.destroy()

    def new_state(self):
        """Add a new state to the machine and return it."""
        s = Node()
        n: cython.Py_ssize_t = self.next_state_number
        self.next_state_number = n + 1
        s.number = n
        self.states.append(s)
        return s

    def new_initial_state(self, name):
        state = self.new_state()
        self.make_initial_state(name, state)
        return state

    def make_initial_state(self, name, state):
        self.initial_states[name] = state

    def get_initial_state(self, name):
        return self.initial_states[name]

    def dump(self, file):
        file.write("Plex.Machine:\n")
        if self.initial_states is not None:
            file.write("   Initial states:\n")
            for (name, state) in sorted(self.initial_states.items()):
                file.write("      '%s': %d\n" % (name, state.number))
        for s in self.states:
            s.dump(file)


class Node:
    """A state of an NFA or DFA."""

    def __init__(self):
        # Preinitialise the list of empty transitions, because
        # the nfa-to-dfa algorithm needs it
        self.transitions = TransitionMap()      # TransitionMap
        self.action_priority = LOWEST_PRIORITY  # integer
        self.action = None  # Action
        self.number = 0     # for debug output
        self.epsilon_closure = None  # used by nfa_to_dfa()

    def destroy(self):
        self.transitions = None
        self.action = None
        self.epsilon_closure = None

    def add_transition(self, event, new_state):
        self.transitions.add(event, new_state)

    def link_to(self, state):
        """Add an epsilon-move from this state to another state."""
        self.add_transition('', state)

    def set_action(self, action, priority):
        """Make this an accepting state with the given action. If
        there is already an action, choose the action with highest
        priority."""
        if priority > self.action_priority:
            self.action = action
            self.action_priority = priority

    def get_action(self):
        return self.action

    def get_action_priority(self):
        return self.action_priority

    def is_accepting(self):
        return self.action is not None

    def __str__(self):
        return "State %d" % self.number

    def dump(self, file):
        # Header
        file.write("   State %d:\n" % self.number)
        # Transitions
        #        self.dump_transitions(file)
        self.transitions.dump(file)
        # Action
        action = self.action
        priority = self.action_priority
        if action is not None:
            file.write("      %s [priority %d]\n" % (action, priority))

    def __lt__(self, other):
        return self.number < other.number

    def __hash__(self):
        # Prevent overflowing hash values due to arbitrarily large unsigned addresses.
        return id(self) & maxint


class FastMachine:
    """
    FastMachine is a deterministic machine represented in a way that
    allows fast scanning.
    """
    def __init__(self):
        self.initial_states = {}  # {state_name:state}
        self.states = []          # [state]  where state = {event:state, 'else':state, 'action':Action}
        self.next_number = 1      # for debugging
        self.new_state_template = {
            '': None, 'bol': None, 'eol': None, 'eof': None, 'else': None
        }

    def __del__(self):
        for state in self.states:
            state.clear()

    def new_state(self, action=None):
        number: cython.Py_ssize_t = self.next_number
        self.next_number = number + 1
        result = self.new_state_template.copy()
        result['number'] = number
        result['action'] = action
        self.states.append(result)
        return result

    def make_initial_state(self, name, state):
        self.initial_states[name] = state

    def add_transitions(self, state: dict, event, new_state, maxint: cython.int = maxint):
        code:  cython.int
        code0: cython.int
        code1: cython.int

        if type(event) is tuple:
            code0, code1 = event
            if code0 == -maxint:
                state['else'] = new_state
            elif code1 != maxint:
                for code in range(code0, code1):
                    state[chr(code)] = new_state
        else:
            state[event] = new_state

    def get_initial_state(self, name):
        return self.initial_states[name]

    def dump(self, file):
        file.write("Plex.FastMachine:\n")
        file.write("   Initial states:\n")
        for name, state in sorted(self.initial_states.items()):
            file.write("      %s: %s\n" % (repr(name), state['number']))
        for state in self.states:
            self.dump_state(state, file)

    def dump_state(self, state, file):
        # Header
        file.write("   State %d:\n" % state['number'])
        # Transitions
        self.dump_transitions(state, file)
        # Action
        action = state['action']
        if action is not None:
            file.write("      %s\n" % action)

    def dump_transitions(self, state, file):
        chars_leading_to_state = {}
        special_to_state = {}
        for (c, s) in state.items():
            if len(c) == 1:
                chars = chars_leading_to_state.get(id(s))
                if chars is None:
                    chars = []
                    chars_leading_to_state[id(s)] = chars
                chars.append(c)
            elif len(c) <= 4:
                special_to_state[c] = s
        ranges_to_state = {}
        for state in self.states:
            char_list = chars_leading_to_state.get(id(state))
            if char_list:
                ranges = self.chars_to_ranges(char_list)
                ranges_to_state[ranges] = state
        for ranges in sorted(ranges_to_state):
            key = self.ranges_to_string(ranges)
            state = ranges_to_state[ranges]
            file.write("      %s --> State %d\n" % (key, state['number']))
        for key in ('bol', 'eol', 'eof', 'else'):
            state = special_to_state.get(key)
            if state:
                file.write("      %s --> State %d\n" % (key, state['number']))

    def chars_to_ranges(self, char_list: list) -> tuple:
        char_list.sort()

        c1: cython.Py_UCS4
        c2: cython.Py_UCS4
        i: cython.Py_ssize_t = 0
        n: cython.Py_ssize_t = len(char_list)
        result = []
        while i < n:
            c1 = ord(char_list[i])
            c2 = c1
            i += 1
            while i < n and ord(char_list[i]) == c2 + 1:
                i += 1
                c2 += 1
            result.append((chr(c1), chr(c2)))
        return tuple(result)

    def ranges_to_string(self, range_list) -> str:
        return ','.join(map(self.range_to_string, range_list))

    def range_to_string(self, range_tuple: tuple):
        (c1, c2) = range_tuple
        if c1 == c2:
            return repr(c1)
        else:
            return f"{c1!r}..{c2!r}"



# Copyright 2014-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno

from portage import os


def first_existing(path):
    """
    Returns the first existing path element, traversing from the given
    path to the root directory. A path is considered to exist if lstat
    either succeeds or raises an error other than ENOENT or ESTALE.

    This can be particularly useful to check if there is permission to
    create a particular file or directory, without actually creating
    anything.

    @param path: a filesystem path
    @type path: str
    @rtype: str
    @return: the element that exists
    """
    existing = False
    for path in iter_parents(path):
        try:
            os.lstat(path)
            existing = True
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                existing = True

        if existing:
            return path

    return os.sep


def iter_parents(path):
    """
    @param path: a filesystem path
    @type path: str
    @rtype: iterator
    @return: an iterator which yields path and all parents of path,
            ending with the root directory
    """
    yield path
    while path != os.sep:
        path = os.path.dirname(path)
        if not path:
            break
        yield path



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import difflib
import re
import portage
from portage import os
from portage.dbapi.porttree import _parse_uri_map
from portage.dbapi.IndexedPortdb import IndexedPortdb
from portage.dbapi.IndexedVardb import IndexedVardb
from portage.exception import InvalidBinaryPackageFormat
from portage.localization import localized_size
from portage.output import bold, darkgreen, green, red
from portage.util import writemsg_stdout
from portage.util.iterators.MultiIterGroupBy import MultiIterGroupBy

from _emerge.Package import Package


class search:
    #
    # class constants
    #
    VERSION_SHORT = 1
    VERSION_RELEASE = 2

    #
    # public interface
    #
    def __init__(
        self,
        root_config,
        spinner,
        searchdesc,
        verbose,
        usepkg,
        usepkgonly,
        search_index=True,
        search_similarity=None,
        fuzzy=True,
        regex_auto=False,
    ):
        """Searches the available and installed packages for the supplied search key.
        The list of available and installed packages is created at object instantiation.
        This makes successive searches faster."""
        self.settings = root_config.settings
        self.verbose = verbose
        self.searchdesc = searchdesc
        self.searchkey = None
        self._results_specified = False
        # Disable the spinner since search results are displayed
        # incrementally.
        self.spinner = None
        self.root_config = root_config
        self.setconfig = root_config.setconfig
        self.regex_auto = regex_auto
        self.fuzzy = fuzzy
        self.search_similarity = 80 if search_similarity is None else search_similarity
        self.matches = {"pkg": []}

        self._dbs = []

        portdb = root_config.trees["porttree"].dbapi
        bindb = root_config.trees["bintree"].dbapi
        vardb = root_config.trees["vartree"].dbapi

        if search_index:
            portdb = IndexedPortdb(portdb)
            vardb = IndexedVardb(vardb)

        if not usepkgonly and portdb._have_root_eclass_dir:
            self._dbs.append(portdb)

        if (usepkg or usepkgonly) and bindb.cp_all():
            self._dbs.append(bindb)

        self._dbs.append(vardb)
        self._portdb = portdb
        self._vardb = vardb

    def _spinner_update(self):
        if self.spinner:
            self.spinner.update()

    def _cp_all(self):
        iterators = []
        for db in self._dbs:
            # MultiIterGroupBy requires sorted input
            i = db.cp_all(sort=True)
            try:
                i = iter(i)
            except TypeError:
                pass
            iterators.append(i)
        for group in MultiIterGroupBy(iterators):
            yield group[0]

    def _aux_get(self, *args, **kwargs):
        for db in self._dbs:
            try:
                return db.aux_get(*args, **kwargs)
            except KeyError:
                pass
        raise KeyError(args[0])

    def _aux_get_error(self, cpv):
        portage.writemsg(
            f"emerge: search: aux_get('{cpv}') failed, skipping\n", noiselevel=-1
        )

    def _findname(self, *args, **kwargs):
        for db in self._dbs:
            if db is not self._portdb:
                # We don't want findname to return anything
                # unless it's an ebuild in a repository.
                # Otherwise, it's already built and we don't
                # care about it.
                continue
            func = getattr(db, "findname", None)
            if func:
                value = func(*args, **kwargs)
                if value:
                    return value
        return None

    def _getFetchMap(self, *args, **kwargs):
        for db in self._dbs:
            func = getattr(db, "getFetchMap", None)
            if func:
                value = func(*args, **kwargs)
                if value:
                    return value
        return {}

    def _visible(self, db, cpv, metadata):
        installed = db is self._vardb
        built = installed or db is not self._portdb
        pkg_type = "ebuild"
        if installed:
            pkg_type = "installed"
        elif built:
            pkg_type = "binary"
        return Package(
            type_name=pkg_type,
            root_config=self.root_config,
            cpv=cpv,
            built=built,
            installed=installed,
            metadata=metadata,
        ).visible

    def _first_cp(self, cp):
        for db in self._dbs:
            if hasattr(db, "cp_list"):
                matches = db.cp_list(cp)
                if matches:
                    return matches[-1]
            else:
                matches = db.match(cp)

            for cpv in matches:
                if cpv.cp == cp:
                    return cpv

        return None

    def _xmatch(self, level, atom):
        """
        This method does not expand old-style virtuals because it
        is restricted to returning matches for a single ${CATEGORY}/${PN}
        and old-style virual matches unreliable for that when querying
        multiple package databases. If necessary, old-style virtuals
        can be performed on atoms prior to calling this method.
        """
        if not isinstance(atom, portage.dep.Atom):
            atom = portage.dep.Atom(atom)

        cp = atom.cp
        if level == "match-all":
            matches = set()
            for db in self._dbs:
                if hasattr(db, "xmatch"):
                    matches.update(db.xmatch(level, atom))
                else:
                    matches.update(db.match(atom))
            result = list(x for x in matches if portage.cpv_getkey(x) == cp)
            db._cpv_sort_ascending(result)
        elif level == "match-visible":
            matches = set()
            for db in self._dbs:
                if hasattr(db, "xmatch"):
                    matches.update(db.xmatch(level, atom))
                else:
                    db_keys = list(db._aux_cache_keys)
                    for cpv in db.match(atom):
                        try:
                            metadata = zip(db_keys, db.aux_get(cpv, db_keys))
                        except KeyError:
                            self._aux_get_error(cpv)
                            continue
                        if not self._visible(db, cpv, metadata):
                            continue
                        matches.add(cpv)
            result = list(x for x in matches if portage.cpv_getkey(x) == cp)
            db._cpv_sort_ascending(result)
        elif level == "bestmatch-visible":
            result = None
            for db in self._dbs:
                if hasattr(db, "xmatch"):
                    cpv = db.xmatch("bestmatch-visible", atom)
                    if not cpv or portage.cpv_getkey(cpv) != cp:
                        continue
                    if not result or cpv == portage.best([cpv, result]):
                        result = cpv
                else:
                    db_keys = list(db._aux_cache_keys)
                    matches = db.match(atom)
                    try:
                        db.match_unordered
                    except AttributeError:
                        pass
                    else:
                        db._cpv_sort_ascending(matches)

                    # break out of this loop with highest visible
                    # match, checked in descending order
                    for cpv in reversed(matches):
                        if portage.cpv_getkey(cpv) != cp:
                            continue
                        try:
                            metadata = zip(db_keys, db.aux_get(cpv, db_keys))
                        except KeyError:
                            self._aux_get_error(cpv)
                            continue
                        if not self._visible(db, cpv, metadata):
                            continue
                        if not result or cpv == portage.best([cpv, result]):
                            result = cpv
                        break
        else:
            raise NotImplementedError(level)
        return result

    def execute(self, searchkey):
        """Performs the search for the supplied search key"""
        self.searchkey = searchkey

    def _iter_search(self):
        match_category = 0
        self.packagematches = []
        if self.searchdesc:
            self.searchdesc = 1
            self.matches = {"pkg": [], "desc": [], "set": []}
        else:
            self.searchdesc = 0
            self.matches = {"pkg": [], "set": []}
        writemsg_stdout("Searching...\n\n", noiselevel=-1)

        regexsearch = False
        if self.searchkey.startswith("%"):
            regexsearch = True
            self.searchkey = self.searchkey[1:]
        if self.searchkey.startswith("@"):
            match_category = 1
            self.searchkey = self.searchkey[1:]
        # Auto-detect category match mode (@ symbol can be deprecated
        # after this is available in a stable version of portage).
        if "/" in self.searchkey:
            match_category = 1
        fuzzy = False

        if (
            self.regex_auto
            and not regexsearch
            and re.search(r"[\^\$\*\[\]\{\}\|\?]|\.\+", self.searchkey) is not None
        ):
            try:
                re.compile(self.searchkey, re.I)
            except Exception:
                pass
            else:
                regexsearch = True

        if regexsearch:
            self.searchre = re.compile(self.searchkey, re.I)
        else:
            self.searchre = re.compile(re.escape(self.searchkey), re.I)

            # Fuzzy search does not support regular expressions, therefore
            # it is disabled for regular expression searches.
            if self.fuzzy:
                fuzzy = True
                cutoff = float(self.search_similarity) / 100
                if match_category:
                    # Weigh the similarity of category and package
                    # names independently, in order to avoid matching
                    # lots of irrelevant packages in the same category
                    # when the package name is much shorter than the
                    # category name.
                    part_split = portage.catsplit
                else:
                    part_split = lambda match_string: (match_string,)

                part_matchers = []
                for part in part_split(self.searchkey):
                    seq_match = difflib.SequenceMatcher()
                    seq_match.set_seq2(part.lower())
                    part_matchers.append(seq_match)

                def fuzzy_search_part(seq_match, match_string):
                    seq_match.set_seq1(match_string.lower())
                    return (
                        seq_match.real_quick_ratio() >= cutoff
                        and seq_match.quick_ratio() >= cutoff
                        and seq_match.ratio() >= cutoff
                    )

                def fuzzy_search(match_string):
                    return all(
                        fuzzy_search_part(seq_match, part)
                        for seq_match, part in zip(
                            part_matchers, part_split(match_string)
                        )
                    )

        for package in self._cp_all():
            self._spinner_update()

            if match_category:
                match_string = package[:]
            else:
                match_string = package.split("/")[-1]

            if self.searchre.search(match_string):
                yield ("pkg", package)
            elif fuzzy and fuzzy_search(match_string):
                yield ("pkg", package)
            elif self.searchdesc:  # DESCRIPTION searching
                # Use _first_cp to avoid an expensive visibility check,
                # since the visibility check can be avoided entirely
                # when the DESCRIPTION does not match.
                full_package = self._first_cp(package)
                if not full_package:
                    continue
                try:
                    full_desc = self._aux_get(full_package, ["DESCRIPTION"])[0]
                except KeyError:
                    self._aux_get_error(full_package)
                    continue
                if not self.searchre.search(full_desc):
                    continue

                yield ("desc", package)

        self.sdict = self.setconfig.getSets()
        for setname in self.sdict:
            self._spinner_update()
            if match_category:
                match_string = setname
            else:
                match_string = setname.split("/")[-1]

            if self.searchre.search(match_string):
                yield ("set", setname)
            elif self.searchdesc:
                if self.searchre.search(self.sdict[setname].getMetadata("DESCRIPTION")):
                    yield ("set", setname)

    def addCP(self, cp):
        """
        Add a specific cp to the search results. This modifies the
        behavior of the output method, so that it only displays specific
        packages added via this method.
        """
        self._results_specified = True
        if not self._xmatch("match-all", cp):
            return
        self.matches["pkg"].append(cp)

    def output(self):
        """Outputs the results of the search."""

        class msg:
            @staticmethod
            def append(msg):
                writemsg_stdout(msg, noiselevel=-1)

        msg.append(
            "\b\b  \n[ Results for search key : " + bold(self.searchkey) + " ]\n"
        )
        vardb = self._vardb
        metadata_keys = set(Package.metadata_keys)
        metadata_keys.update(["DESCRIPTION", "HOMEPAGE", "LICENSE", "SRC_URI"])
        metadata_keys = tuple(metadata_keys)

        if self._results_specified:
            # Handle results added via addCP
            addCP_matches = []
            for mytype, matches in self.matches.items():
                for match in matches:
                    addCP_matches.append((mytype, match))
            iterator = iter(addCP_matches)

        else:
            # Do a normal search
            iterator = self._iter_search()

        mlen = 0
        for mtype, match in iterator:
            mlen += 1
            masked = False
            full_package = None
            if mtype in ("pkg", "desc"):
                full_package = self._xmatch("bestmatch-visible", match)
                if not full_package:
                    masked = True
                    full_package = self._xmatch("match-all", match)
                    if full_package:
                        full_package = full_package[-1]
            elif mtype == "set":
                msg.append(green("*") + "  " + bold(match) + "\n")
                if self.verbose:
                    msg.append(
                        "      "
                        + darkgreen("Description:")
                        + "   "
                        + self.sdict[match].getMetadata("DESCRIPTION")
                        + "\n\n"
                    )
            if full_package:
                try:
                    metadata = dict(
                        zip(metadata_keys, self._aux_get(full_package, metadata_keys))
                    )
                except KeyError:
                    self._aux_get_error(full_package)
                    continue

                desc = metadata["DESCRIPTION"]
                homepage = metadata["HOMEPAGE"]
                license = metadata["LICENSE"]  # pylint: disable=redefined-builtin

                if masked:
                    msg.append(
                        green("*") + "  " + bold(match) + " " + red("[ Masked ]") + "\n"
                    )
                else:
                    msg.append(green("*") + "  " + bold(match) + "\n")
                myversion = self.getVersion(full_package, search.VERSION_RELEASE)

                mysum = [0, 0]
                file_size_str = None
                mycat = match.split("/")[0]
                mypkg = match.split("/")[1]
                mycpv = match + "-" + myversion
                myebuild = self._findname(mycpv)
                if myebuild:
                    pkg = Package(
                        built=False,
                        cpv=mycpv,
                        installed=False,
                        metadata=metadata,
                        root_config=self.root_config,
                        type_name="ebuild",
                    )
                    pkgdir = os.path.dirname(myebuild)
                    mf = self.settings.repositories.get_repo_for_location(
                        os.path.dirname(os.path.dirname(pkgdir))
                    )
                    mf = mf.load_manifest(pkgdir, self.settings["DISTDIR"])
                    try:
                        uri_map = _parse_uri_map(mycpv, metadata, use=pkg.use.enabled)
                    except portage.exception.InvalidDependString as e:
                        file_size_str = f"Unknown ({e})"
                        del e
                    else:
                        try:
                            mysum[0] = mf.getDistfilesSize(uri_map)
                        except KeyError as e:
                            file_size_str = "Unknown (missing " + f"digest for {e})"
                            del e

                available = False
                for db in self._dbs:
                    if db is not vardb and db.cpv_exists(mycpv):
                        available = True
                        if not myebuild and hasattr(db, "bintree"):
                            try:
                                myebuild = db.bintree.getname(mycpv)
                            except InvalidBinaryPackageFormat:
                                break
                            try:
                                mysum[0] = os.stat(myebuild).st_size
                            except OSError:
                                myebuild = None
                        break

                if myebuild and file_size_str is None:
                    file_size_str = localized_size(mysum[0])

                if self.verbose:
                    if available:
                        msg.append(
                            f"      {darkgreen('Latest version available:')} {myversion}\n"
                        )
                    msg.append(
                        f"      {self.getInstallationStatus(mycat + '/' + mypkg)}\n"
                    )
                    if myebuild:
                        msg.append(
                            f"      {darkgreen('Size of files:')} {file_size_str}\n"
                        )
                    msg.append(
                        "      " + darkgreen("Homepage:") + "      " + homepage + "\n"
                    )
                    msg.append(
                        "      " + darkgreen("Description:") + "   " + desc + "\n"
                    )
                    msg.append(
                        "      " + darkgreen("License:") + "       " + license + "\n\n"
                    )

        msg.append("[ Applications found : " + bold(str(mlen)) + " ]\n\n")

    #
    # private interface
    #
    def getInstallationStatus(self, package):
        if not isinstance(package, portage.dep.Atom):
            package = portage.dep.Atom(package)

        installed_package = self._vardb.match(package)
        if installed_package:
            try:
                self._vardb.match_unordered
            except AttributeError:
                installed_package = installed_package[-1]
            else:
                installed_package = portage.best(installed_package)

        else:
            installed_package = ""
        result = ""
        version = self.getVersion(installed_package, search.VERSION_RELEASE)
        if len(version) > 0:
            result = darkgreen("Latest version installed:") + " " + version
        else:
            result = darkgreen("Latest version installed:") + " [ Not Installed ]"
        return result

    def getVersion(self, full_package, detail):
        if len(full_package) > 1:
            package_parts = portage.catpkgsplit(full_package)
            if detail == search.VERSION_RELEASE and package_parts[3] != "r0":
                result = package_parts[2] + "-" + package_parts[3]
            else:
                result = package_parts[2]
        else:
            result = ""
        return result



# Copyright 2018-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os

from portage.tests import TestCase
from portage.util._async.AsyncFunction import AsyncFunction
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio


def fork_main():
    loop = asyncio._wrap_loop()
    # Before python 3.12 this fails with python's default event loop policy,
    # see https://bugs.python.org/issue22087.
    loop.run_until_complete(asyncio.sleep(0.1, loop=loop))
    loop.close()


def async_main(fork_exitcode, loop=None):
    loop = asyncio._wrap_loop(loop)
    proc = AsyncFunction(scheduler=loop, target=fork_main)
    proc.start()
    proc.async_wait().add_done_callback(
        lambda future: fork_exitcode.set_result(future.result())
    )


class EventLoopInForkTestCase(TestCase):
    """
    Before python 3.12 the default asyncio event loop policy does not support loops
    running in forks, see https://bugs.python.org/issue22087.
    """

    def testEventLoopInForkTestCase(self):
        loop = global_event_loop()
        fork_exitcode = loop.create_future()
        # Make async_main fork while the loop is running, which would
        # trigger https://bugs.python.org/issue22087 with asyncio's
        # default event loop policy before python 3.12.
        loop.call_soon(async_main, fork_exitcode)
        assert loop.run_until_complete(fork_exitcode) == os.EX_OK



# Copyright 2012-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SlotAbiTestCase(TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def testSubSlot(self):
        ebuilds = {
            "dev-libs/icu-49": {"EAPI": "5", "SLOT": "0/49"},
            "dev-libs/icu-4.8": {"EAPI": "5", "SLOT": "0/48"},
            "dev-libs/libxml2-2.7.8": {
                "EAPI": "5",
                "DEPEND": "dev-libs/icu:=",
                "RDEPEND": "dev-libs/icu:=",
            },
        }
        binpkgs = {
            "dev-libs/icu-49": {"EAPI": "5", "SLOT": "0/49"},
            "dev-libs/icu-4.8": {"EAPI": "5", "SLOT": "0/48"},
            "dev-libs/libxml2-2.7.8": {
                "EAPI": "5",
                "DEPEND": "dev-libs/icu:0/48=",
                "RDEPEND": "dev-libs/icu:0/48=",
            },
        }
        installed = {
            "dev-libs/icu-4.8": {"EAPI": "5", "SLOT": "0/48"},
            "dev-libs/libxml2-2.7.8": {
                "EAPI": "5",
                "DEPEND": "dev-libs/icu:0/48=",
                "RDEPEND": "dev-libs/icu:0/48=",
            },
        }

        world = ["dev-libs/libxml2"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True},
                success=True,
                mergelist=["dev-libs/icu-49", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True, "--ignore-built-slot-operator-deps": "y"},
                success=True,
                mergelist=["dev-libs/icu-49"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/icu-49", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True, "--usepkgonly": True},
                success=True,
                mergelist=["[binary]dev-libs/icu-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={
                    "--oneshot": True,
                    "--usepkgonly": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["[binary]dev-libs/icu-49"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-libs/icu-49", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["dev-libs/icu-49"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/icu-49", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkgonly": True},
                success=True,
                mergelist=[],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--usepkgonly": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["[binary]dev-libs/icu-49"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": (f'BINPKG_FORMAT="{binpkg_format}"',),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()

    def testWholeSlot(self):
        ebuilds = {
            "sys-libs/db-4.8": {"SLOT": "4.8"},
            "sys-libs/db-4.7": {"SLOT": "4.7"},
            "app-office/libreoffice-3.5.4.2": {
                "EAPI": "5",
                "DEPEND": ">=sys-libs/db-4:=",
                "RDEPEND": ">=sys-libs/db-4:=",
            },
        }
        binpkgs = {
            "sys-libs/db-4.8": {"SLOT": "4.8"},
            "sys-libs/db-4.7": {"SLOT": "4.7"},
            "app-office/libreoffice-3.5.4.2": {
                "EAPI": "5",
                "DEPEND": ">=sys-libs/db-4:4.7/4.7=",
                "RDEPEND": ">=sys-libs/db-4:4.7/4.7=",
            },
        }
        installed = {
            "sys-libs/db-4.7": {"SLOT": "4.7"},
            "app-office/libreoffice-3.5.4.2": {
                "EAPI": "5",
                "DEPEND": ">=sys-libs/db-4:4.7/4.7=",
                "RDEPEND": ">=sys-libs/db-4:4.7/4.7=",
            },
        }

        world = ["app-office/libreoffice"]

        test_cases = (
            # The first 2 test cases don't trigger a libreoffice rebuild
            # because sys-libs/db is the only package requested, and a
            # rebuild is not necessary because the sys-libs/db:4.7 slot
            # remains installed.
            ResolverPlaygroundTestCase(
                ["sys-libs/db"],
                options={"--oneshot": True},
                success=True,
                mergelist=["sys-libs/db-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["sys-libs/db"],
                options={"--oneshot": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]sys-libs/db-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["sys-libs/db"],
                options={"--oneshot": True, "--usepkgonly": True},
                success=True,
                mergelist=["[binary]sys-libs/db-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["sys-libs/db"],
                options={"--oneshot": True, "--rebuild-if-new-slot": "n"},
                success=True,
                mergelist=["sys-libs/db-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["sys-libs/db-4.8", "app-office/libreoffice-3.5.4.2"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]sys-libs/db-4.8", "app-office/libreoffice-3.5.4.2"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--usepkg": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["[binary]sys-libs/db-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkgonly": True},
                success=True,
                mergelist=[],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--usepkgonly": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["[binary]sys-libs/db-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--rebuild-if-new-slot": "n",
                },
                success=True,
                mergelist=[],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": (f'BINPKG_FORMAT="{binpkg_format}"',),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()

    def testWholeSlotConditional(self):
        ebuilds = {
            "dev-libs/libnl-3.2.14": {"SLOT": "3"},
            "dev-libs/libnl-1.1-r3": {"SLOT": "1.1"},
            "net-misc/networkmanager-0.9.6.4-r1": {
                "EAPI": "5",
                "IUSE": "wimax",
                "DEPEND": "wimax? ( dev-libs/libnl:1.1= ) !wimax? ( dev-libs/libnl:3= )",
                "RDEPEND": "wimax? ( dev-libs/libnl:1.1= ) !wimax? ( dev-libs/libnl:3= )",
            },
        }
        installed = {
            "dev-libs/libnl-1.1-r3": {"SLOT": "1.1"},
            "net-misc/networkmanager-0.9.6.4-r1": {
                "EAPI": "5",
                "IUSE": "wimax",
                "USE": "wimax",
                "DEPEND": "dev-libs/libnl:1.1/1.1=",
                "RDEPEND": "dev-libs/libnl:1.1/1.1=",
            },
        }

        user_config = {"make.conf": ('USE="wimax"',)}

        world = ["net-misc/networkmanager"]

        test_cases = (
            # Demonstrate bug #460304, where _slot_operator_update_probe needs
            # to account for USE conditional deps.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            user_config=user_config,
            world=world,
            debug=False,
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

        user_config = {"make.conf": ('USE="-wimax"',)}

        test_cases = (
            # Demonstrate bug #460304 again, but with inverted USE
            # settings this time.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "dev-libs/libnl-3.2.14",
                    "net-misc/networkmanager-0.9.6.4-r1",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            user_config=user_config,
            world=world,
            debug=False,
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testWholeSlotSubSlotMix(self):
        ebuilds = {
            "dev-libs/glib-1.2.10": {"SLOT": "1"},
            "dev-libs/glib-2.30.2": {"EAPI": "5", "SLOT": "2/2.30"},
            "dev-libs/glib-2.32.3": {"EAPI": "5", "SLOT": "2/2.32"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2=",
                "RDEPEND": "dev-libs/glib:2=",
            },
        }
        binpkgs = {
            "dev-libs/glib-1.2.10": {"SLOT": "1"},
            "dev-libs/glib-2.30.2": {"EAPI": "5", "SLOT": "2/2.30"},
            "dev-libs/glib-2.32.3": {"EAPI": "5", "SLOT": "2/2.32"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2/2.30=",
                "RDEPEND": "dev-libs/glib:2/2.30=",
            },
        }
        installed = {
            "dev-libs/glib-1.2.10": {"EAPI": "5", "SLOT": "1"},
            "dev-libs/glib-2.30.2": {"EAPI": "5", "SLOT": "2/2.30"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2/2.30=",
                "RDEPEND": "dev-libs/glib:2/2.30=",
            },
        }

        world = ["dev-libs/glib:1", "dev-libs/dbus-glib"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True},
                success=True,
                mergelist=["dev-libs/glib-2.32.3", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True, "--ignore-built-slot-operator-deps": "y"},
                success=True,
                mergelist=["dev-libs/glib-2.32.3"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/glib-2.32.3", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True, "--usepkgonly": True},
                success=True,
                mergelist=["[binary]dev-libs/glib-2.30.2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={
                    "--oneshot": True,
                    "--usepkgonly": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["[binary]dev-libs/glib-2.32.3"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-libs/glib-2.32.3", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["dev-libs/glib-2.32.3"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/glib-2.32.3", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkgonly": True},
                success=True,
                mergelist=[],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--usepkgonly": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["[binary]dev-libs/glib-2.32.3"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": (f'BINPKG_FORMAT="{binpkg_format}"',),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



from libcpp cimport bool

cdef extern from "<optional>" namespace "std" nogil:
    cdef cppclass nullopt_t:
        nullopt_t()

    cdef nullopt_t nullopt

    cdef cppclass optional[T]:
        ctypedef T value_type
        optional()
        optional(nullopt_t)
        optional(optional&) except +
        optional(T&) except +
        bool has_value()
        T& value() except +
        T& value_or[U](U& default_value)
        void swap(optional&)
        void reset()
        T& emplace(...)
        T& operator*()
        #T* operator->() # Not Supported
        optional& operator=(optional&)
        optional& operator=[U](U&)
        bool operator bool()
        bool operator!()
        bool operator==[U](optional&, U&)
        bool operator!=[U](optional&, U&)
        bool operator<[U](optional&, U&)
        bool operator>[U](optional&, U&)
        bool operator<=[U](optional&, U&)
        bool operator>=[U](optional&, U&)

    optional[T] make_optional[T](...) except +



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SimpleResolverTestCase(TestCase):
    def testSimple(self):
        ebuilds = {
            "dev-libs/A-1": {"KEYWORDS": "x86"},
            "dev-libs/A-2": {"KEYWORDS": "~x86"},
            "dev-libs/B-1.2": {},
            "app-misc/Z-1": {
                "DEPEND": "|| ( app-misc/Y ( app-misc/X app-misc/W ) )",
                "RDEPEND": "",
            },
            "app-misc/Y-1": {"KEYWORDS": "~x86"},
            "app-misc/X-1": {},
            "app-misc/W-1": {},
        }
        binpkgs = {
            "dev-libs/B-1.2": {},
        }
        installed = {
            "dev-libs/A-1": {},
            "dev-libs/B-1.1": {},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A"], success=True, mergelist=["dev-libs/A-1"]
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-2"], options={"--autounmask": "n"}, success=False
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--noreplace": True},
                success=True,
                mergelist=[],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--noreplace": True},
                success=True,
                mergelist=[],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--update": True},
                success=True,
                mergelist=["dev-libs/B-1.2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--update": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/B-1.2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--update": True, "--usepkgonly": True},
                success=True,
                mergelist=["[binary]dev-libs/B-1.2"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/Z"],
                success=True,
                ambiguous_merge_order=True,
                mergelist=[("app-misc/W-1", "app-misc/X-1"), "app-misc/Z-1"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    user_config={
                        "make.conf": (f'BINPKG_FORMAT="{binpkg_format}"',),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# Copyright 2001-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


# The format for a tbz2/xpak:
#
#  tbz2: tar.bz2 + xpak + (xpak_offset) + "STOP"
#  xpak: "XPAKPACK" + (index_len) + (data_len) + index + data + "XPAKSTOP"
# index: (pathname_len) + pathname + (data_offset) + (data_len)
#        index entries are concatenated end-to-end.
#  data: concatenated data chunks, end-to-end.
#
# [tarball]XPAKPACKIIIIDDDD[index][data]XPAKSTOPOOOOSTOP
#
# (integer) == encodeint(integer)  ===> 4 characters (big-endian copy)
# '+' means concatenate the fields ===> All chunks are strings

__all__ = [
    "addtolist",
    "decodeint",
    "encodeint",
    "getboth",
    "getindex",
    "getindex_mem",
    "getitem",
    "listindex",
    "searchindex",
    "tbz2",
    "xpak_mem",
    "xpak",
    "xpand",
    "xsplit",
    "xsplit_mem",
]

import array
import errno
import tempfile

import portage
from portage import os
from portage import shutil
from portage import normalize_path
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.binpkg import get_binpkg_format
from portage.exception import InvalidBinaryPackageFormat
from portage.util.file_copy import copyfile


def addtolist(mylist, curdir):
    """(list, dir) --- Takes an array(list) and appends all files from dir down
    the directory tree. Returns nothing. list is modified."""
    curdir = normalize_path(
        _unicode_decode(curdir, encoding=_encodings["fs"], errors="strict")
    )
    for parent, dirs, files in os.walk(curdir):
        parent = _unicode_decode(parent, encoding=_encodings["fs"], errors="strict")
        if parent != curdir:
            mylist.append(parent[len(curdir) + 1 :] + os.sep)

        for x in dirs:
            try:
                _unicode_decode(x, encoding=_encodings["fs"], errors="strict")
            except UnicodeDecodeError:
                dirs.remove(x)

        for x in files:
            try:
                x = _unicode_decode(x, encoding=_encodings["fs"], errors="strict")
            except UnicodeDecodeError:
                continue
            mylist.append(os.path.join(parent, x)[len(curdir) + 1 :])


def encodeint(myint):
    """Takes a 4 byte integer and converts it into a string of 4 characters.
    Returns the characters in a string."""
    a = array.array("B")
    a.append((myint >> 24) & 0xFF)
    a.append((myint >> 16) & 0xFF)
    a.append((myint >> 8) & 0xFF)
    a.append(myint & 0xFF)
    try:
        # Python >= 3.2
        return a.tobytes()
    except AttributeError:
        return a.tostring()


def decodeint(mystring):
    """Takes a 4 byte string and converts it into a 4 byte integer.
    Returns an integer."""
    myint = 0
    myint += mystring[3]
    myint += mystring[2] << 8
    myint += mystring[1] << 16
    myint += mystring[0] << 24
    return myint


def xpak(rootdir, outfile=None):
    """(rootdir, outfile) -- creates an xpak segment of the directory 'rootdir'
    and under the name 'outfile' if it is specified. Otherwise it returns the
    xpak segment."""

    if portage.utf8_mode and not isinstance(rootdir, bytes):
        # Since paths are encoded below, rootdir must also be encoded
        # when _unicode_func_wrapper is not used.
        rootdir = os.fsencode(rootdir)

    mylist = []

    addtolist(mylist, rootdir)
    mylist.sort()
    mydata = {}
    for x in mylist:
        if x == "CONTENTS":
            # CONTENTS is generated during the merge process.
            continue
        x = _unicode_encode(x, encoding=_encodings["fs"], errors="strict")
        with open(os.path.join(rootdir, x), "rb") as f:
            mydata[x] = f.read()

    xpak_segment = xpak_mem(mydata)
    if outfile:
        outf = open(
            _unicode_encode(outfile, encoding=_encodings["fs"], errors="strict"), "wb"
        )
        outf.write(xpak_segment)
        outf.close()
    else:
        return xpak_segment


def xpak_mem(mydata):
    """Create an xpack segment from a map object."""

    mydata_encoded = {}
    for k, v in mydata.items():
        k = _unicode_encode(
            k, encoding=_encodings["repo.content"], errors="backslashreplace"
        )
        v = _unicode_encode(
            v, encoding=_encodings["repo.content"], errors="backslashreplace"
        )
        mydata_encoded[k] = v
    mydata = mydata_encoded
    del mydata_encoded

    indexglob = b""
    indexpos = 0
    dataglob = b""
    datapos = 0
    for x, newglob in mydata.items():
        mydatasize = len(newglob)
        indexglob = (
            indexglob
            + encodeint(len(x))
            + x
            + encodeint(datapos)
            + encodeint(mydatasize)
        )
        indexpos = indexpos + 4 + len(x) + 4 + 4
        dataglob = dataglob + newglob
        datapos = datapos + mydatasize
    return (
        b"XPAKPACK"
        + encodeint(len(indexglob))
        + encodeint(len(dataglob))
        + indexglob
        + dataglob
        + b"XPAKSTOP"
    )


def xsplit(infile):
    """(infile) -- Splits the infile into two files.
    'infile.index' contains the index segment.
    'infile.dat' contains the data segment."""
    infile = _unicode_decode(infile, encoding=_encodings["fs"], errors="strict")
    myfile = open(
        _unicode_encode(infile, encoding=_encodings["fs"], errors="strict"), "rb"
    )
    mydat = myfile.read()
    myfile.close()

    splits = xsplit_mem(mydat)
    if not splits:
        return False

    myfile = open(
        _unicode_encode(infile + ".index", encoding=_encodings["fs"], errors="strict"),
        "wb",
    )
    myfile.write(splits[0])
    myfile.close()
    myfile = open(
        _unicode_encode(infile + ".dat", encoding=_encodings["fs"], errors="strict"),
        "wb",
    )
    myfile.write(splits[1])
    myfile.close()
    return True


def xsplit_mem(mydat):
    if mydat[0:8] != b"XPAKPACK":
        return None
    if mydat[-8:] != b"XPAKSTOP":
        return None
    indexsize = decodeint(mydat[8:12])
    return (mydat[16 : indexsize + 16], mydat[indexsize + 16 : -8])


def getindex(infile):
    """(infile) -- grabs the index segment from the infile and returns it."""
    myfile = open(
        _unicode_encode(infile, encoding=_encodings["fs"], errors="strict"), "rb"
    )
    myheader = myfile.read(16)
    if myheader[0:8] != b"XPAKPACK":
        myfile.close()
        return
    indexsize = decodeint(myheader[8:12])
    myindex = myfile.read(indexsize)
    myfile.close()
    return myindex


def getboth(infile):
    """(infile) -- grabs the index and data segments from the infile.
    Returns an array [indexSegment, dataSegment]"""
    myfile = open(
        _unicode_encode(infile, encoding=_encodings["fs"], errors="strict"), "rb"
    )
    myheader = myfile.read(16)
    if myheader[0:8] != b"XPAKPACK":
        myfile.close()
        return
    indexsize = decodeint(myheader[8:12])
    datasize = decodeint(myheader[12:16])
    myindex = myfile.read(indexsize)
    mydata = myfile.read(datasize)
    myfile.close()
    return myindex, mydata


def listindex(myindex):
    """Print to the terminal the filenames listed in the indexglob passed in."""
    for x in getindex_mem(myindex):
        print(x)


def getindex_mem(myindex):
    """Returns the filenames listed in the indexglob passed in."""
    myindexlen = len(myindex)
    startpos = 0
    myret = []
    while (startpos + 8) < myindexlen:
        mytestlen = decodeint(myindex[startpos : startpos + 4])
        myret = myret + [myindex[startpos + 4 : startpos + 4 + mytestlen]]
        startpos = startpos + mytestlen + 12
    return myret


def searchindex(myindex, myitem):
    """(index, item) -- Finds the offset and length of the file 'item' in the
    datasegment via the index 'index' provided."""
    myitem = _unicode_encode(
        myitem, encoding=_encodings["repo.content"], errors="backslashreplace"
    )
    mylen = len(myitem)
    myindexlen = len(myindex)
    startpos = 0
    while (startpos + 8) < myindexlen:
        mytestlen = decodeint(myindex[startpos : startpos + 4])
        if mytestlen == mylen:
            if myitem == myindex[startpos + 4 : startpos + 4 + mytestlen]:
                # found
                datapos = decodeint(
                    myindex[startpos + 4 + mytestlen : startpos + 8 + mytestlen]
                )
                datalen = decodeint(
                    myindex[startpos + 8 + mytestlen : startpos + 12 + mytestlen]
                )
                return datapos, datalen
        startpos = startpos + mytestlen + 12


def getitem(myid, myitem):
    myindex = myid[0]
    mydata = myid[1]
    myloc = searchindex(myindex, myitem)
    if not myloc:
        return None
    return mydata[myloc[0] : myloc[0] + myloc[1]]


def xpand(myid, mydest):
    mydest = normalize_path(mydest) + os.sep
    myindex = myid[0]
    mydata = myid[1]
    myindexlen = len(myindex)
    startpos = 0
    while (startpos + 8) < myindexlen:
        namelen = decodeint(myindex[startpos : startpos + 4])
        datapos = decodeint(myindex[startpos + 4 + namelen : startpos + 8 + namelen])
        datalen = decodeint(myindex[startpos + 8 + namelen : startpos + 12 + namelen])
        myname = myindex[startpos + 4 : startpos + 4 + namelen]
        myname = _unicode_decode(
            myname, encoding=_encodings["repo.content"], errors="replace"
        )
        filename = os.path.join(mydest, myname.lstrip(os.sep))
        filename = normalize_path(filename)
        if not filename.startswith(mydest):
            # myname contains invalid ../ component(s)
            continue
        dirname = os.path.dirname(filename)
        if dirname:
            if not os.path.exists(dirname):
                os.makedirs(dirname)
        mydat = open(
            _unicode_encode(filename, encoding=_encodings["fs"], errors="strict"), "wb"
        )
        mydat.write(mydata[datapos : datapos + datalen])
        mydat.close()
        startpos = startpos + namelen + 12


class tbz2:
    def __init__(self, myfile):
        self.file = myfile
        self.filestat = None
        self.index = b""
        self.infosize = 0
        self.xpaksize = 0
        self.indexsize = None
        self.datasize = None
        self.indexpos = None
        self.datapos = None

    def decompose(self, datadir, cleanup=1):
        """Alias for unpackinfo() --- Complement to recompose() but optionally
        deletes the destination directory. Extracts the xpak from the tbz2 into
        the directory provided. Raises IOError if scan() fails.
        Returns result of upackinfo()."""
        if not self.scan():
            raise OSError
        if cleanup:
            self.cleanup(datadir)
        if not os.path.exists(datadir):
            os.makedirs(datadir)
        return self.unpackinfo(datadir)

    def compose(self, datadir, cleanup=0):
        """Alias for recompose()."""
        return self.recompose(datadir, cleanup)

    def recompose(self, datadir, cleanup=0, break_hardlinks=True):
        """Creates an xpak segment from the datadir provided, truncates the tbz2
        to the end of regular data if an xpak segment already exists, and adds
        the new segment to the file with terminating info."""
        xpdata = xpak(datadir)
        self.recompose_mem(xpdata, break_hardlinks=break_hardlinks)
        if cleanup:
            self.cleanup(datadir)

    def recompose_mem(self, xpdata, break_hardlinks=True):
        """
        Update the xpak segment.
        @param xpdata: A new xpak segment to be written, like that returned
                from the xpak_mem() function.
        @param break_hardlinks: If hardlinks exist, create a copy in order
                to break them. This makes it safe to use hardlinks to create
                cheap snapshots of the repository, which is useful for solving
                race conditions on binhosts as described here:
                https://crbug.com/185031
                Default is True.
        """
        self.scan()  # Don't care about condition... We'll rewrite the data anyway.

        if break_hardlinks and self.filestat and self.filestat.st_nlink > 1:
            with tempfile.NamedTemporaryFile(
                dir=os.path.dirname(self.file),
                prefix=f"{os.path.basename(self.file)}.{portage.getpid()}",
            ) as safe_temp:
                tmp_fname = safe_temp.name
            copyfile(self.file, tmp_fname)
            try:
                portage.util.apply_stat_permissions(self.file, self.filestat)
            except portage.exception.OperationNotPermitted:
                pass
            os.rename(tmp_fname, self.file)

        myfile = open(
            _unicode_encode(self.file, encoding=_encodings["fs"], errors="strict"),
            "ab+",
        )
        if not myfile:
            raise OSError
        myfile.seek(-self.xpaksize, 2)  # 0,2 or -0,2 just mean EOF.
        myfile.truncate()
        myfile.write(xpdata + encodeint(len(xpdata)) + b"STOP")
        myfile.flush()
        myfile.close()
        return 1

    def cleanup(self, datadir):
        datadir_split = os.path.split(datadir)
        if len(datadir_split) >= 2 and len(datadir_split[1]) > 0:
            # This is potentially dangerous,
            # thus the above sanity check.
            try:
                shutil.rmtree(datadir)
            except OSError as oe:
                if oe.errno == errno.ENOENT:
                    pass
                else:
                    raise oe

    def scan(self):
        """Scans the tbz2 to locate the xpak segment and setup internal values.
        This function is called by relevant functions already."""
        a = None
        try:
            mystat = os.stat(self.file)
            if self.filestat:
                changed = 0
                if (
                    mystat.st_size != self.filestat.st_size
                    or mystat.st_mtime != self.filestat.st_mtime
                    or mystat.st_ctime != self.filestat.st_ctime
                ):
                    changed = True
                if not changed:
                    return 1
            self.filestat = mystat
            a = open(
                _unicode_encode(self.file, encoding=_encodings["fs"], errors="strict"),
                "rb",
            )
            a.seek(-16, 2)
            trailer = a.read()
            self.infosize = 0
            self.xpaksize = 0
            if trailer[-4:] != b"STOP":
                try:
                    get_binpkg_format(self.file, check_file=True)
                except InvalidBinaryPackageFormat:
                    pass
                return 0
            if trailer[0:8] != b"XPAKSTOP":
                try:
                    get_binpkg_format(self.file, check_file=True)
                except InvalidBinaryPackageFormat:
                    pass
                return 0
            self.infosize = decodeint(trailer[8:12])
            self.xpaksize = self.infosize + 8
            a.seek(-(self.xpaksize), 2)
            header = a.read(16)
            if header[0:8] != b"XPAKPACK":
                try:
                    get_binpkg_format(self.file, check_file=True)
                except InvalidBinaryPackageFormat:
                    pass
                return 0
            self.indexsize = decodeint(header[8:12])
            self.datasize = decodeint(header[12:16])
            self.indexpos = a.tell()
            self.index = a.read(self.indexsize)
            self.datapos = a.tell()
            return 2
        except SystemExit:
            raise
        except:
            try:
                get_binpkg_format(self.file, check_file=True)
            except InvalidBinaryPackageFormat:
                pass
            return 0
        finally:
            if a is not None:
                a.close()

    def filelist(self):
        """Return an array of each file listed in the index."""
        if not self.scan():
            return None
        return getindex_mem(self.index)

    def getfile(self, myfile, mydefault=None):
        """Finds 'myfile' in the data segment and returns it."""
        if not self.scan():
            return None
        myresult = searchindex(self.index, myfile)
        if not myresult:
            return mydefault
        a = open(
            _unicode_encode(self.file, encoding=_encodings["fs"], errors="strict"), "rb"
        )
        a.seek(self.datapos + myresult[0], 0)
        myreturn = a.read(myresult[1])
        a.close()
        return myreturn

    def getelements(self, myfile):
        """A split/array representation of tbz2.getfile()"""
        mydat = self.getfile(myfile)
        if not mydat:
            return []
        return mydat.split()

    def unpackinfo(self, mydest):
        """Unpacks all the files from the dataSegment into 'mydest'."""
        if not self.scan():
            return 0
        mydest = normalize_path(mydest) + os.sep
        a = open(
            _unicode_encode(self.file, encoding=_encodings["fs"], errors="strict"), "rb"
        )
        if not os.path.exists(mydest):
            os.makedirs(mydest)
        startpos = 0
        while (startpos + 8) < self.indexsize:
            namelen = decodeint(self.index[startpos : startpos + 4])
            datapos = decodeint(
                self.index[startpos + 4 + namelen : startpos + 8 + namelen]
            )
            datalen = decodeint(
                self.index[startpos + 8 + namelen : startpos + 12 + namelen]
            )
            myname = self.index[startpos + 4 : startpos + 4 + namelen]
            myname = _unicode_decode(
                myname, encoding=_encodings["repo.content"], errors="replace"
            )
            filename = os.path.join(mydest, myname.lstrip(os.sep))
            filename = normalize_path(filename)
            if not filename.startswith(mydest):
                # myname contains invalid ../ component(s)
                continue
            dirname = os.path.dirname(filename)
            if dirname:
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
            mydat = open(
                _unicode_encode(filename, encoding=_encodings["fs"], errors="strict"),
                "wb",
            )
            a.seek(self.datapos + datapos)
            mydat.write(a.read(datalen))
            mydat.close()
            startpos = startpos + namelen + 12
        a.close()
        return 1

    def get_data(self):
        """Returns all the files from the dataSegment as a map object."""
        if not self.scan():
            return {}
        a = open(
            _unicode_encode(self.file, encoding=_encodings["fs"], errors="strict"), "rb"
        )
        mydata = {}
        startpos = 0
        while (startpos + 8) < self.indexsize:
            namelen = decodeint(self.index[startpos : startpos + 4])
            datapos = decodeint(
                self.index[startpos + 4 + namelen : startpos + 8 + namelen]
            )
            datalen = decodeint(
                self.index[startpos + 8 + namelen : startpos + 12 + namelen]
            )
            myname = self.index[startpos + 4 : startpos + 4 + namelen]
            a.seek(self.datapos + datapos)
            mydata[myname] = a.read(datalen)
            startpos = startpos + namelen + 12
        a.close()
        return mydata

    def getboth(self):
        """Returns an array [indexSegment, dataSegment]"""
        if not self.scan():
            return None

        a = open(
            _unicode_encode(self.file, encoding=_encodings["fs"], errors="strict"), "rb"
        )
        a.seek(self.datapos)
        mydata = a.read(self.datasize)
        a.close()

        return self.index, mydata



cdef extern from "Python.h":

    ###########################################################################
    # MemoryView Objects
    ###########################################################################
    # A memoryview object exposes the C level buffer interface as a Python
    # object which can then be passed around like any other object

    memoryview PyMemoryView_FromObject(object obj)
    # Return value: New reference.
    # Create a memoryview object from an object that provides the buffer
    # interface. If obj supports writable buffer exports, the memoryview object
    # will be read/write, otherwise it may be either read-only or read/write at
    # the discretion of the exporter.

    memoryview PyMemoryView_FromMemory(char *mem, Py_ssize_t size, int flags)
    # Return value: New reference.
    # Create a memoryview object using mem as the underlying buffer. flags can
    # be one of PyBUF_READ or PyBUF_WRITE.
    # New in version 3.3.

    memoryview PyMemoryView_FromBuffer(Py_buffer *view)
    # Return value: New reference.
    # Create a memoryview object wrapping the given buffer structure view. For
    # simple byte buffers, PyMemoryView_FromMemory() is the preferred function.

    memoryview PyMemoryView_GetContiguous(object obj,
                                          int buffertype,
                                          char order)
    # Return value: New reference.
    # Create a memoryview object to a contiguous chunk of memory (in either ‘C’
    # or ‘F’ortran order) from an object that defines the buffer interface. If
    # memory is contiguous, the memoryview object points to the original
    # memory. Otherwise, a copy is made and the memoryview points to a new
    # bytes object.

    bint PyMemoryView_Check(object obj)
    # Return true if the object obj is a memoryview object. It is not currently
    # allowed to create subclasses of memoryview.

    Py_buffer *PyMemoryView_GET_BUFFER(object mview)
    # Return a pointer to the memoryview’s private copy of the exporter’s
    # buffer. mview must be a memoryview instance; this macro doesn’t check its
    # type, you must do it yourself or you will risk crashes.

    Py_buffer *PyMemoryView_GET_BASE(object mview)
    # Return either a pointer to the exporting object that the memoryview is
    # based on or NULL if the memoryview has been created by one of the
    # functions PyMemoryView_FromMemory() or PyMemoryView_FromBuffer(). mview
    # must be a memoryview instance.



import textwrap
from unittest import TestCase

from .. import Naming
from ..Code import _indent_chunk, UtilityCode, process_utility_ccode

class TestIndent(TestCase):
    def _test_indentations(self, chunk, expected):
        for indentation in range(16):
            expected_indented = textwrap.indent(expected, ' ' * indentation)
            for line in expected_indented.splitlines():
                # Validate before the comparison that empty lines got stripped also by textwrap.indent().
                self.assertTrue(line == '' or line.strip(), repr(line))

            with self.subTest(indentation=indentation):
                result = _indent_chunk(chunk, indentation_length=indentation)
                self.assertEqual(expected_indented, result)

    def test_indent_empty(self):
        self._test_indentations('', '')

    def test_indent_empty_lines(self):
        self._test_indentations('\n', '\n')
        self._test_indentations('\n'*2, '\n'*2)
        self._test_indentations('\n'*3, '\n'*3)
        self._test_indentations(' \n'*2, '\n'*2)
        self._test_indentations('\n  \n \n    \n', '\n'*4)

    def test_indent_one_line(self):
        self._test_indentations('abc', 'abc')

    def test_indent_chunk(self):
        chunk = """
            x = 1
            if x == 2:
                print("False")
            else:
                print("True")
        """
        expected = """
x = 1
if x == 2:
    print("False")
else:
    print("True")
"""
        self._test_indentations(chunk, expected)

    def test_indent_empty_line(self):
        chunk = """
            x = 1

            if x == 2:
                print("False")
            else:
                print("True")
        """
        expected = """
x = 1

if x == 2:
    print("False")
else:
    print("True")
"""
        self._test_indentations(chunk, expected)

    def test_indent_empty_line_unclean(self):
        lines = """
            x = 1

            if x == 2:
                print("False")
            else:
                print("True")
        """.splitlines(keepends=True)
        lines[2] = '            \n'
        chunk = ''.join(lines)
        expected = """
x = 1

if x == 2:
    print("False")
else:
    print("True")
"""
        self._test_indentations(chunk, expected)


class TestUtilityCodeProcessing(TestCase):
    def _process(self, code):
        utility_code = UtilityCode()
        formatted_code, is_module_specific = process_utility_ccode(utility_code, None, code)
        self.assertFalse(is_module_specific)  # cannot currently test this case
        return formatted_code

    def assert_formatted_code(self, code: str, expected: str, dedent=False):
        if dedent:
            expected = textwrap.dedent(expected)
        expected = expected.strip() + '\n\n'
        formatted = self._process(code)
        self.assertEqual(formatted, expected)

    def test_format_cstring(self):
        self.assert_formatted_code('''
        Some Text and a CSTRING("""
        spanning "multiple" 'lines'.
        Really.
        """);   # end of C string
        ''',
        expected=r'''
        Some Text and a "\n"
        "        spanning \042multiple\042 'lines'.\n"
        "        Really.\n"
        "        \n"
        ;   # end of C string
        ''',
        dedent=True)

    def test_cglobal(self):
        self.assert_formatted_code("""
        CGLOBAL(name)
        NAMED_CGLOBAL(empty_tuple)
        """,
        expected=f"""
        {Naming.modulestateglobal_cname}->name
        {Naming.modulestateglobal_cname}->{Naming.empty_tuple}
        """)

    def test_empty_builtin(self):
        self.assert_formatted_code("""
        EMPTY(tuple)EMPTY(bytes)
        EMPTY(tuple);EMPTY(bytes)
        EMPTY(unicode)
        EMPTY(bytes)
        EMPTY(tuple)
        """,
        expected=f"""
        {Naming.modulestateglobal_cname}->{Naming.empty_tuple}{Naming.modulestateglobal_cname}->{Naming.empty_bytes}
        {Naming.modulestateglobal_cname}->{Naming.empty_tuple};{Naming.modulestateglobal_cname}->{Naming.empty_bytes}
        {Naming.modulestateglobal_cname}->{Naming.empty_unicode}
        {Naming.modulestateglobal_cname}->{Naming.empty_bytes}
        {Naming.modulestateglobal_cname}->{Naming.empty_tuple}
        """)



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class CompeteIfNewSubSlotWithoutRevBumpTestCase(TestCase):
    def testCompeteIfNewSubSlotWithoutRevBump(self):
        ebuilds = {
            "media-libs/libpng-1.5.14": {"EAPI": "5", "SLOT": "0"},
            "x11-libs/gdk-pixbuf-2.26.5": {
                "EAPI": "5",
                "DEPEND": ">=media-libs/libpng-1.4:=",
                "RDEPEND": ">=media-libs/libpng-1.4:=",
            },
        }

        binpkgs = {
            "x11-libs/gdk-pixbuf-2.26.5": {
                "EAPI": "5",
                "DEPEND": ">=media-libs/libpng-1.4:0/15=",
                "RDEPEND": ">=media-libs/libpng-1.4:0/15=",
            },
        }

        installed = {
            "media-libs/libpng-1.5.14": {"EAPI": "5", "SLOT": "0/15"},
            "x11-libs/gdk-pixbuf-2.26.5": {
                "EAPI": "5",
                "DEPEND": ">=media-libs/libpng-1.4:0/15=",
                "RDEPEND": ">=media-libs/libpng-1.4:0/15=",
            },
        }

        world = ["x11-libs/gdk-pixbuf"]

        test_cases = (
            # Test that --complete-graph-if-new-ver=y triggers rebuild
            # when the sub-slot changes without a revbump.
            ResolverPlaygroundTestCase(
                ["media-libs/libpng"],
                options={
                    "--oneshot": True,
                    "--complete-graph-if-new-ver": "y",
                    "--rebuild-if-new-slot": "n",
                    "--usepkg": True,
                },
                success=True,
                mergelist=["media-libs/libpng-1.5.14", "x11-libs/gdk-pixbuf-2.26.5"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": (f'BINPKG_FORMAT="{binpkg_format}"',),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



import argparse

from certifi import contents, where, __version__


parser = argparse.ArgumentParser(prog="certifi-system-store")
parser.add_argument("-c", "--contents", action="store_true")
parser.add_argument("-v", "--verbose", action="store_true")
parser.add_argument("--system-store", action="store_true")
args = parser.parse_args()

if args.verbose:
    print(f"certifi-system store {__version__}")

if args.contents:
    print(contents())
else:
    print(where())



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util._dyn_libs.NeededEntry import NeededEntry
from portage.util._dyn_libs.soname_deps import SonameDepsProcessor


class SonameDepsProcessorTestCase(TestCase):
    def testInternalLibsWithoutSoname(self):
        """
        Test handling of internal libraries that lack an soname, which are
        resolved via DT_RUNPATH, see ebtables for example (bug 646190).
        """
        needed_elf_2 = """
X86_64;/sbin/ebtables;;/lib64/ebtables;libebt_802_3.so,libebtable_broute.so,libc.so.6;x86_64
X86_64;/lib64/ebtables/libebtable_broute.so;;;libc.so.6;x86_64
X86_64;/lib64/ebtables/libebt_802_3.so;;;libc.so.6;x86_64
"""
        soname_deps = SonameDepsProcessor("", "")

        for line in needed_elf_2.splitlines():
            if not line:
                continue
            entry = NeededEntry.parse(None, line)
            soname_deps.add(entry)

        self.assertEqual(soname_deps.provides, None)
        # Prior to the fix for bug 646190, REQUIRES contained references to
        # the internal libebt* libraries which are resolved via a DT_RUNPATH
        # entry referring to the /lib64/ebtables directory that contains the
        # internal libraries.
        self.assertEqual(soname_deps.requires, "x86_64: libc.so.6\n")



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021-2025 Intel Corporation
# Copyright © 2021-2025 Intel Corporation
from __future__ import annotations

"""Keyword Argument type annotations."""

import typing as T

from typing_extensions import TypedDict, Literal, Protocol, NotRequired

from .. import build
from .. import options
from ..compilers import Compiler
from ..dependencies.base import Dependency
from ..mesonlib import EnvironmentVariables, MachineChoice, File, FileMode, FileOrString
from ..options import OptionKey
from ..modules.cmake import CMakeSubprojectOptions
from ..programs import ExternalProgram
from .type_checking import PkgConfigDefineType, SourcesVarargsType

if T.TYPE_CHECKING:
    TestArgs = T.Union[str, File, build.Target, ExternalProgram]

class FuncAddProjectArgs(TypedDict):

    """Keyword Arguments for the add_*_arguments family of arguments.

    including `add_global_arguments`, `add_project_arguments`, and their
    link variants

    Because of the use of a convertor function, we get the native keyword as
    a MachineChoice instance already.
    """

    native: MachineChoice
    language: T.List[str]


class BaseTest(TypedDict):

    """Shared base for the Rust module."""

    should_fail: bool
    timeout: int
    workdir: T.Optional[str]
    depends: T.List[T.Union[build.CustomTarget, build.BuildTarget]]
    priority: int
    env: EnvironmentVariables
    suite: T.List[str]


class FuncBenchmark(BaseTest):

    """Keyword Arguments shared between `test` and `benchmark`."""

    args: T.List[TestArgs]
    protocol: Literal['exitcode', 'tap', 'gtest', 'rust']


class FuncTest(FuncBenchmark):

    """Keyword Arguments for `test`

    `test` only adds the `is_parallel` argument over benchmark, so inheritance
    is helpful here.
    """

    is_parallel: bool


class ExtractRequired(TypedDict):

    """Keyword Arguments consumed by the `extract_required_kwargs` function.

    Any function that uses the `required` keyword argument which accepts either
    a boolean or a feature option should inherit its arguments from this class.
    """

    required: T.Union[bool, options.UserFeatureOption]


class ExtractSearchDirs(TypedDict):

    """Keyword arguments consumed by the `extract_search_dirs` function.

    See the not in `ExtractRequired`
    """

    dirs: T.List[str]


class FuncGenerator(TypedDict):

    """Keyword rguments for the generator function."""

    arguments: T.List[str]
    output: T.List[str]
    depfile: T.Optional[str]
    capture:  bool
    depends: T.List[T.Union[build.BuildTarget, build.CustomTarget]]


class GeneratorProcess(TypedDict):

    """Keyword Arguments for generator.process."""

    preserve_path_from: T.Optional[str]
    extra_args: T.List[str]
    env: EnvironmentVariables

class DependencyMethodPartialDependency(TypedDict):

    """ Keyword Arguments for the dep.partial_dependency methods """

    compile_args: bool
    link_args: bool
    links: bool
    includes: bool
    sources: bool

class BuildTargeMethodExtractAllObjects(TypedDict):
    recursive: bool

class FuncInstallSubdir(TypedDict):

    install_dir: str
    strip_directory: bool
    exclude_files: T.List[str]
    exclude_directories: T.List[str]
    install_mode: FileMode
    follow_symlinks: T.Optional[bool]


class FuncInstallData(TypedDict):

    install_dir: str
    sources: T.List[FileOrString]
    rename: T.List[str]
    install_mode: FileMode
    follow_symlinks: T.Optional[bool]


class FuncInstallHeaders(TypedDict):

    install_dir: T.Optional[str]
    install_mode: FileMode
    subdir: T.Optional[str]
    follow_symlinks: T.Optional[bool]


class FuncInstallMan(TypedDict):

    install_dir: T.Optional[str]
    install_mode: FileMode
    locale: T.Optional[str]


class FuncImportModule(ExtractRequired):

    disabler: bool


class FuncIncludeDirectories(TypedDict):

    is_system: bool

class FuncAddLanguages(ExtractRequired):

    native: T.Optional[bool]

class RunTarget(TypedDict):

    command: T.List[T.Union[str, build.BuildTarget, build.CustomTarget, ExternalProgram, File]]
    depends: T.List[T.Union[build.BuildTarget, build.CustomTarget]]
    env: EnvironmentVariables


class CustomTarget(TypedDict):

    build_always: bool
    build_always_stale: T.Optional[bool]
    build_by_default: T.Optional[bool]
    capture: bool
    command: T.List[T.Union[str, build.BuildTarget, build.CustomTarget,
                            build.CustomTargetIndex, ExternalProgram, File]]
    console: bool
    depend_files: T.List[FileOrString]
    depends: T.List[T.Union[build.BuildTarget, build.CustomTarget]]
    depfile: T.Optional[str]
    env: EnvironmentVariables
    feed: bool
    input: T.List[T.Union[str, build.BuildTarget, build.CustomTarget, build.CustomTargetIndex,
                          build.ExtractedObjects, build.GeneratedList, ExternalProgram, File]]
    install: bool
    install_dir: T.List[T.Union[str, T.Literal[False]]]
    install_mode: FileMode
    install_tag: T.List[T.Optional[str]]
    output: T.List[str]

class AddTestSetup(TypedDict):

    exe_wrapper: T.List[T.Union[str, ExternalProgram]]
    gdb: bool
    timeout_multiplier: int
    is_default: bool
    exclude_suites: T.List[str]
    env: EnvironmentVariables


class Project(TypedDict):

    version: T.Optional[FileOrString]
    meson_version: T.Optional[str]
    default_options: T.List[str]
    license: T.List[str]
    license_files: T.List[str]
    subproject_dir: str


class _FoundProto(Protocol):

    """Protocol for subdir arguments.

    This allows us to define any object that has a found(self) -> bool method
    """

    def found(self) -> bool: ...


class Subdir(TypedDict):

    if_found: T.List[_FoundProto]


class Summary(TypedDict):

    section: str
    bool_yn: bool
    list_sep: T.Optional[str]


class FindProgram(ExtractRequired, ExtractSearchDirs):

    default_options: T.Dict[OptionKey, options.ElementaryOptionValues]
    native: MachineChoice
    version: T.List[str]


class RunCommand(TypedDict):

    check: bool
    capture: T.Optional[bool]
    env: EnvironmentVariables


class FeatureOptionRequire(TypedDict):

    error_message: T.Optional[str]


class DependencyPkgConfigVar(TypedDict):

    default: T.Optional[str]
    define_variable: PkgConfigDefineType


class DependencyGetVariable(TypedDict):

    cmake: T.Optional[str]
    pkgconfig: T.Optional[str]
    configtool: T.Optional[str]
    internal: T.Optional[str]
    system: T.Optional[str]
    default_value: T.Optional[str]
    pkgconfig_define: PkgConfigDefineType


class ConfigurationDataSet(TypedDict):

    description: T.Optional[str]

class VcsTag(TypedDict):

    command: T.List[T.Union[str, build.BuildTarget, build.CustomTarget,
                            build.CustomTargetIndex, ExternalProgram, File]]
    fallback: T.Optional[str]
    input: T.List[T.Union[str, build.BuildTarget, build.CustomTarget, build.CustomTargetIndex,
                          build.ExtractedObjects, build.GeneratedList, ExternalProgram, File]]
    output: T.List[str]
    replace_string: str
    install: bool
    install_tag: T.Optional[str]
    install_dir: T.Optional[str]
    install_mode: FileMode


class ConfigureFile(TypedDict):

    output: str
    capture: bool
    format: T.Literal['meson', 'cmake', 'cmake@']
    output_format: T.Literal['c', 'json', 'nasm']
    depfile: T.Optional[str]
    install: T.Optional[bool]
    install_dir: T.Union[str, T.Literal[False]]
    install_mode: FileMode
    install_tag: T.Optional[str]
    encoding: str
    command: T.Optional[T.List[T.Union[build.Executable, ExternalProgram, Compiler, File, str]]]
    input: T.List[FileOrString]
    configuration: T.Optional[T.Union[T.Dict[str, T.Union[str, int, bool]], build.ConfigurationData]]
    macro_name: T.Optional[str]


class Subproject(ExtractRequired):

    default_options: T.Dict[OptionKey, options.ElementaryOptionValues]
    version: T.List[str]


class DoSubproject(ExtractRequired):

    default_options: T.Union[T.List[str], T.Dict[str, options.ElementaryOptionValues], str]
    version: T.List[str]
    cmake_options: T.List[str]
    options: T.Optional[CMakeSubprojectOptions]


class _BaseBuildTarget(TypedDict):

    """Arguments used by all BuildTarget like functions.

    This really exists because Jar is so different than all of the other
    BuildTarget functions.
    """

    build_by_default: bool
    build_rpath: str
    extra_files: T.List[FileOrString]
    gnu_symbol_visibility: str
    install: bool
    install_mode: FileMode
    install_rpath: str
    implicit_include_directories: bool
    link_depends: T.List[T.Union[str, File, build.CustomTarget, build.CustomTargetIndex, build.BuildTarget]]
    link_language: T.Optional[str]
    name_prefix: T.Optional[str]
    name_suffix: T.Optional[str]
    native: MachineChoice
    objects: T.List[build.ObjectTypes]
    override_options: T.Dict[OptionKey, options.ElementaryOptionValues]
    depend_files: NotRequired[T.List[File]]
    resources: T.List[str]


class _BuildTarget(_BaseBuildTarget):

    """Arguments shared by non-JAR functions"""

    d_debug: T.List[T.Union[str, int]]
    d_import_dirs: T.List[T.Union[str, build.IncludeDirs]]
    d_module_versions: T.List[T.Union[str, int]]
    d_unittest: bool
    rust_dependency_map: T.Dict[str, str]
    swift_interoperability_mode: Literal['c', 'cpp']
    swift_module_name: str
    sources: SourcesVarargsType
    c_args: T.List[str]
    cpp_args: T.List[str]
    cuda_args: T.List[str]
    fortran_args: T.List[str]
    d_args: T.List[str]
    objc_args: T.List[str]
    objcpp_args: T.List[str]
    rust_args: T.List[str]
    vala_args: T.List[T.Union[str, File]]  # Yes, Vala is really special
    cs_args: T.List[str]
    swift_args: T.List[str]
    cython_args: T.List[str]
    nasm_args: T.List[str]
    masm_args: T.List[str]


class _LibraryMixin(TypedDict):

    rust_abi: T.Optional[Literal['c', 'rust']]


class Executable(_BuildTarget):

    export_dynamic: T.Optional[bool]
    gui_app: T.Optional[bool]
    implib: T.Optional[T.Union[str, bool]]
    pie: T.Optional[bool]
    vs_module_defs: T.Optional[T.Union[str, File, build.CustomTarget, build.CustomTargetIndex]]
    win_subsystem: T.Optional[str]
    android_exe_type: T.Optional[Literal['application', 'executable']]


class _StaticLibMixin(TypedDict):

    prelink: bool
    pic: T.Optional[bool]


class StaticLibrary(_BuildTarget, _StaticLibMixin, _LibraryMixin):
    pass


class _SharedLibMixin(TypedDict):

    darwin_versions: T.Optional[T.Tuple[str, str]]
    soversion: T.Optional[str]
    version: T.Optional[str]
    vs_module_defs: T.Optional[T.Union[str, File, build.CustomTarget, build.CustomTargetIndex]]


class SharedLibrary(_BuildTarget, _SharedLibMixin, _LibraryMixin):
    pass


class SharedModule(_BuildTarget, _LibraryMixin):

    vs_module_defs: T.Optional[T.Union[str, File, build.CustomTarget, build.CustomTargetIndex]]


class Library(_BuildTarget, _SharedLibMixin, _StaticLibMixin, _LibraryMixin):

    """For library, both_library, and as a base for build_target"""

    c_static_args: NotRequired[T.List[str]]
    c_shared_args: NotRequired[T.List[str]]
    cpp_static_args: NotRequired[T.List[str]]
    cpp_shared_args: NotRequired[T.List[str]]
    cuda_static_args: NotRequired[T.List[str]]
    cuda_shared_args: NotRequired[T.List[str]]
    fortran_static_args: NotRequired[T.List[str]]
    fortran_shared_args: NotRequired[T.List[str]]
    d_static_args: NotRequired[T.List[str]]
    d_shared_args: NotRequired[T.List[str]]
    objc_static_args: NotRequired[T.List[str]]
    objc_shared_args: NotRequired[T.List[str]]
    objcpp_static_args: NotRequired[T.List[str]]
    objcpp_shared_args: NotRequired[T.List[str]]
    rust_static_args: NotRequired[T.List[str]]
    rust_shared_args: NotRequired[T.List[str]]
    vala_static_args: NotRequired[T.List[T.Union[str, File]]]  # Yes, Vala is really special
    vala_shared_args: NotRequired[T.List[T.Union[str, File]]]  # Yes, Vala is really special
    cs_static_args: NotRequired[T.List[str]]
    cs_shared_args: NotRequired[T.List[str]]
    swift_static_args: NotRequired[T.List[str]]
    swift_shared_args: NotRequired[T.List[str]]
    cython_static_args: NotRequired[T.List[str]]
    cython_shared_args: NotRequired[T.List[str]]
    nasm_static_args: NotRequired[T.List[str]]
    nasm_shared_args: NotRequired[T.List[str]]
    masm_static_args: NotRequired[T.List[str]]
    masm_shared_args: NotRequired[T.List[str]]


class BuildTarget(Library):

    target_type: Literal['executable', 'shared_library', 'static_library',
                         'shared_module', 'both_libraries', 'library', 'jar']


class Jar(_BaseBuildTarget):

    main_class: str
    java_resources: T.Optional[build.StructuredSources]
    sources: T.Union[str, File, build.CustomTarget, build.CustomTargetIndex, build.GeneratedList, build.ExtractedObjects, build.BuildTarget]
    java_args: T.List[str]


class FuncDeclareDependency(TypedDict):

    compile_args: T.List[str]
    d_import_dirs: T.List[T.Union[build.IncludeDirs, str]]
    d_module_versions: T.List[T.Union[str, int]]
    dependencies: T.List[Dependency]
    extra_files: T.List[FileOrString]
    include_directories: T.List[T.Union[build.IncludeDirs, str]]
    link_args: T.List[str]
    link_whole: T.List[T.Union[build.StaticLibrary, build.CustomTarget, build.CustomTargetIndex]]
    link_with: T.List[build.LibTypes]
    objects: T.List[build.ExtractedObjects]
    sources: T.List[T.Union[FileOrString, build.GeneratedTypes]]
    variables: T.Dict[str, str]
    version: T.Optional[str]


class FuncDependency(TypedDict):

    default_options: T.Dict[OptionKey, T.Union[str, int, bool, T.List[str]]]



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class PerlRebuildBugTestCase(TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def testPerlRebuildBug(self):
        """
        The infamous Perl rebuild bug.

        A non-slotted build-time dependency cycle is created by:
        dev-lang/perl -> sys-libs/zlib -> sys-devel/automake -> dev-lang/perl
        Everything else depends on this cycle.

        Bug in solving for smallest cycle causes slot in RDEPEND of
        dev-perl/Locale-gettext to be ignored, so all dependencies other than
        perl's >=sys-libs/zlib-1.2.12 are satisfied by already-installed
        packages. dev-perl/Locale-gettext and sys-devel/automake become leaves
        of the depgraph after satisfied packages are ignored. They become
        emerged first. This causes an issue because dev-perl/Locale-gettext is
        now built before the slot upgrade of dev-lang/perl.
        """
        ebuilds = {
            "dev-lang/perl-5.36.0-r2": {
                "EAPI": "5",
                "DEPEND": ">=sys-libs/zlib-1.2.12",
                "RDEPEND": ">=sys-libs/zlib-1.2.12",
                "SLOT": "0/5.36",
            },
            "dev-perl/Locale-gettext-1.70.0-r1": {
                "EAPI": "5",
                "DEPEND": "dev-lang/perl",
                "RDEPEND": "dev-lang/perl:=",
            },
            "sys-apps/help2man-1.49.3": {
                "EAPI": "5",
                "DEPEND": "dev-lang/perl dev-perl/Locale-gettext",
                "RDEPEND": "dev-lang/perl dev-perl/Locale-gettext",
            },
            "sys-devel/automake-1.16.5": {
                "EAPI": "5",
                "DEPEND": "dev-lang/perl",
                "RDEPEND": "dev-lang/perl",
            },
            "sys-libs/zlib-1.2.13-r1": {
                "EAPI": "5",
                "DEPEND": "sys-devel/automake",
            },
        }

        installed = {
            "dev-lang/perl-5.34.0-r3": {
                "EAPI": "5",
                "DEPEND": "sys-libs/zlib",
                "RDEPEND": "sys-libs/zlib",
                "SLOT": "0/5.34",
            },
            "dev-perl/Locale-gettext-1.70.0-r1": {
                "EAPI": "5",
                "DEPEND": "dev-lang/perl",
                "RDEPEND": "dev-lang/perl:0/5.34=",
            },
            "sys-apps/help2man-1.48.5": {
                "EAPI": "5",
                "DEPEND": "dev-lang/perl dev-perl/Locale-gettext",
                "RDEPEND": "dev-lang/perl dev-perl/Locale-gettext",
            },
            "sys-devel/automake-1.16.4": {
                "EAPI": "5",
                "DEPEND": "dev-lang/perl",
                "RDEPEND": "dev-lang/perl",
            },
            "sys-libs/zlib-1.2.11-r4": {
                "EAPI": "5",
                "DEPEND": "sys-devel/automake",
            },
        }

        world = ["sys-apps/help2man"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--deep": True, "--update": True, "--verbose": True},
                success=True,
                ambiguous_merge_order=True,
                merge_order_assertions=(
                    (
                        "dev-lang/perl-5.36.0-r2",
                        "dev-perl/Locale-gettext-1.70.0-r1",
                    ),
                ),
                mergelist=[
                    "sys-devel/automake-1.16.5",
                    "sys-libs/zlib-1.2.13-r1",
                    "dev-lang/perl-5.36.0-r2",
                    "dev-perl/Locale-gettext-1.70.0-r1",
                    "sys-apps/help2man-1.49.3",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            world=world,
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2011-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class MergeOrderTestCase(TestCase):
    def testMergeOrder(self):
        ebuilds = {
            "app-misc/blocker-buildtime-a-1": {},
            "app-misc/blocker-buildtime-unbuilt-a-1": {
                "DEPEND": "!app-misc/installed-blocker-a",
            },
            "app-misc/blocker-buildtime-unbuilt-hard-a-1": {
                "EAPI": "2",
                "DEPEND": "!!app-misc/installed-blocker-a",
            },
            "app-misc/blocker-update-order-a-1": {},
            "app-misc/blocker-update-order-hard-a-1": {},
            "app-misc/blocker-update-order-hard-unsolvable-a-1": {},
            "app-misc/blocker-runtime-a-1": {},
            "app-misc/blocker-runtime-b-1": {},
            "app-misc/blocker-runtime-hard-a-1": {},
            "app-misc/circ-buildtime-a-0": {},
            "app-misc/circ-buildtime-a-1": {
                "RDEPEND": "app-misc/circ-buildtime-b",
            },
            "app-misc/circ-buildtime-b-1": {
                "RDEPEND": "app-misc/circ-buildtime-c",
            },
            "app-misc/circ-buildtime-c-1": {
                "DEPEND": "app-misc/circ-buildtime-a",
            },
            "app-misc/circ-buildtime-unsolvable-a-1": {
                "RDEPEND": "app-misc/circ-buildtime-unsolvable-b",
            },
            "app-misc/circ-buildtime-unsolvable-b-1": {
                "RDEPEND": "app-misc/circ-buildtime-unsolvable-c",
            },
            "app-misc/circ-buildtime-unsolvable-c-1": {
                "DEPEND": "app-misc/circ-buildtime-unsolvable-a",
            },
            "app-misc/circ-post-runtime-a-1": {
                "PDEPEND": "app-misc/circ-post-runtime-b",
            },
            "app-misc/circ-post-runtime-b-1": {
                "RDEPEND": "app-misc/circ-post-runtime-c",
            },
            "app-misc/circ-post-runtime-c-1": {
                "RDEPEND": "app-misc/circ-post-runtime-a",
            },
            "app-misc/circ-runtime-a-1": {
                "RDEPEND": "app-misc/circ-runtime-b",
            },
            "app-misc/circ-runtime-b-1": {
                "RDEPEND": "app-misc/circ-runtime-c",
            },
            "app-misc/circ-runtime-c-1": {
                "RDEPEND": "app-misc/circ-runtime-a",
            },
            "app-misc/circ-satisfied-a-0": {
                "RDEPEND": "app-misc/circ-satisfied-b",
            },
            "app-misc/circ-satisfied-a-1": {
                "RDEPEND": "app-misc/circ-satisfied-b",
            },
            "app-misc/circ-satisfied-b-0": {
                "RDEPEND": "app-misc/circ-satisfied-c",
            },
            "app-misc/circ-satisfied-b-1": {
                "RDEPEND": "app-misc/circ-satisfied-c",
            },
            "app-misc/circ-satisfied-c-0": {
                "DEPEND": "app-misc/circ-satisfied-a",
                "RDEPEND": "app-misc/circ-satisfied-a",
            },
            "app-misc/circ-satisfied-c-1": {
                "DEPEND": "app-misc/circ-satisfied-a",
                "RDEPEND": "app-misc/circ-satisfied-a",
            },
            "app-misc/circ-direct-a-1": {
                "RDEPEND": "app-misc/circ-direct-b",
            },
            "app-misc/circ-direct-b-1": {
                "RDEPEND": "app-misc/circ-direct-a",
                "DEPEND": "app-misc/circ-direct-a",
            },
            "app-misc/circ-smallest-a-1": {
                "RDEPEND": "app-misc/circ-smallest-b",
            },
            "app-misc/circ-smallest-b-1": {
                "RDEPEND": "app-misc/circ-smallest-a",
            },
            "app-misc/circ-smallest-c-1": {
                "RDEPEND": "app-misc/circ-smallest-d",
            },
            "app-misc/circ-smallest-d-1": {
                "RDEPEND": "app-misc/circ-smallest-e",
            },
            "app-misc/circ-smallest-e-1": {
                "RDEPEND": "app-misc/circ-smallest-c",
            },
            "app-misc/circ-smallest-f-1": {
                "RDEPEND": "app-misc/circ-smallest-g app-misc/circ-smallest-a app-misc/circ-smallest-c",
            },
            "app-misc/circ-smallest-g-1": {
                "RDEPEND": "app-misc/circ-smallest-f",
            },
            "app-misc/installed-blocker-a-1": {
                "EAPI": "2",
                "DEPEND": "!app-misc/blocker-buildtime-a",
                "RDEPEND": "!app-misc/blocker-runtime-a !app-misc/blocker-runtime-b !!app-misc/blocker-runtime-hard-a",
            },
            "app-misc/installed-old-version-blocks-a-1": {
                "RDEPEND": "!app-misc/blocker-update-order-a",
            },
            "app-misc/installed-old-version-blocks-a-2": {},
            "app-misc/installed-old-version-blocks-hard-a-1": {
                "EAPI": "2",
                "RDEPEND": "!!app-misc/blocker-update-order-hard-a",
            },
            "app-misc/installed-old-version-blocks-hard-a-2": {},
            "app-misc/installed-old-version-blocks-hard-unsolvable-a-1": {
                "EAPI": "2",
                "RDEPEND": "!!app-misc/blocker-update-order-hard-unsolvable-a",
            },
            "app-misc/installed-old-version-blocks-hard-unsolvable-a-2": {
                "DEPEND": "app-misc/blocker-update-order-hard-unsolvable-a",
                "RDEPEND": "",
            },
            "app-misc/some-app-a-1": {
                "RDEPEND": "app-misc/circ-runtime-a app-misc/circ-runtime-b",
            },
            "app-misc/some-app-b-1": {
                "RDEPEND": "app-misc/circ-post-runtime-a app-misc/circ-post-runtime-b",
            },
            "app-misc/some-app-c-1": {
                "RDEPEND": "app-misc/circ-buildtime-a app-misc/circ-buildtime-b",
            },
            "app-admin/eselect-python-20100321": {},
            "sys-apps/portage-2.1.9.42": {
                "DEPEND": "dev-lang/python",
                "RDEPEND": "dev-lang/python",
            },
            "sys-apps/portage-2.1.9.49": {
                "DEPEND": "dev-lang/python >=app-admin/eselect-python-20091230",
                "RDEPEND": "dev-lang/python",
            },
            "dev-lang/python-3.1": {},
            "dev-lang/python-3.2": {},
            "virtual/libc-0": {
                "RDEPEND": "sys-libs/glibc",
            },
            "sys-devel/gcc-4.5.2": {},
            "sys-devel/binutils-2.18": {},
            "sys-devel/binutils-2.20.1": {},
            "sys-libs/glibc-2.11": {
                "DEPEND": "virtual/os-headers sys-devel/gcc sys-devel/binutils",
                "RDEPEND": "",
            },
            "sys-libs/glibc-2.13": {
                "DEPEND": "virtual/os-headers sys-devel/gcc sys-devel/binutils",
                "RDEPEND": "",
            },
            "virtual/os-headers-0": {
                "RDEPEND": "sys-kernel/linux-headers",
            },
            "sys-kernel/linux-headers-2.6.38": {
                "DEPEND": "app-arch/xz-utils",
                "RDEPEND": "",
            },
            "sys-kernel/linux-headers-2.6.39": {
                "DEPEND": "app-arch/xz-utils",
                "RDEPEND": "",
            },
            "app-arch/xz-utils-5.0.1": {},
            "app-arch/xz-utils-5.0.2": {},
            "dev-util/pkgconfig-0.25-r2": {},
            "kde-base/kdelibs-3.5.7": {
                "PDEPEND": "kde-misc/kdnssd-avahi",
            },
            "kde-misc/kdnssd-avahi-0.1.2": {
                "DEPEND": "kde-base/kdelibs app-arch/xz-utils dev-util/pkgconfig",
                "RDEPEND": "kde-base/kdelibs",
            },
            "kde-base/kdnssd-3.5.7": {
                "DEPEND": "kde-base/kdelibs",
                "RDEPEND": "kde-base/kdelibs",
            },
            "kde-base/libkdegames-3.5.7": {
                "DEPEND": "kde-base/kdelibs",
                "RDEPEND": "kde-base/kdelibs",
            },
            "kde-base/kmines-3.5.7": {
                "DEPEND": "kde-base/libkdegames",
                "RDEPEND": "kde-base/libkdegames",
            },
            "media-libs/mesa-9.1.3": {
                "EAPI": "5",
                "IUSE": "+xorg",
                "DEPEND": "xorg? ( x11-base/xorg-server:= )",
                "RDEPEND": "xorg? ( x11-base/xorg-server:= )",
            },
            "media-video/libav-0.7_pre20110327": {
                "EAPI": "2",
                "IUSE": "X +encode",
                "RDEPEND": "!media-video/ffmpeg",
            },
            "media-video/ffmpeg-0.7_rc1": {
                "EAPI": "2",
                "IUSE": "X +encode",
            },
            "virtual/ffmpeg-0.6.90": {
                "EAPI": "2",
                "IUSE": "X +encode",
                "RDEPEND": "|| ( >=media-video/ffmpeg-0.6.90_rc0-r2[X=,encode=] >=media-video/libav-0.6.90_rc[X=,encode=] )",
            },
            "x11-base/xorg-drivers-1.20-r2": {
                "EAPI": "7",
                "IUSE": "+video_cards_fbdev",
                "PDEPEND": "x11-base/xorg-server video_cards_fbdev? ( x11-drivers/xf86-video-fbdev )",
            },
            "x11-base/xorg-server-1.14.1": {
                "EAPI": "5",
                "SLOT": "0/1.14.1",
                "DEPEND": "media-libs/mesa",
                "RDEPEND": "media-libs/mesa",
                "PDEPEND": "x11-base/xorg-drivers",
            },
            "x11-drivers/xf86-video-fbdev-0.5.0-r1": {
                "EAPI": "7",
                "DEPEND": "x11-base/xorg-server",
                "RDEPEND": "x11-base/xorg-server:=",
            },
        }

        installed = {
            "app-misc/circ-direct-a-1": {
                "RDEPEND": "app-misc/circ-direct-b",
            },
            "app-misc/circ-direct-b-1": {
                "RDEPEND": "app-misc/circ-direct-a",
                "DEPEND": "app-misc/circ-direct-a",
            },
            "app-misc/circ-buildtime-a-0": {},
            "app-misc/circ-satisfied-a-0": {
                "RDEPEND": "app-misc/circ-satisfied-b",
            },
            "app-misc/circ-satisfied-b-0": {
                "RDEPEND": "app-misc/circ-satisfied-c",
            },
            "app-misc/circ-satisfied-c-0": {
                "DEPEND": "app-misc/circ-satisfied-a",
                "RDEPEND": "app-misc/circ-satisfied-a",
            },
            "app-misc/installed-blocker-a-1": {
                "EAPI": "2",
                "DEPEND": "!app-misc/blocker-buildtime-a",
                "RDEPEND": "!app-misc/blocker-runtime-a !app-misc/blocker-runtime-b !!app-misc/blocker-runtime-hard-a",
            },
            "app-misc/installed-old-version-blocks-a-1": {
                "RDEPEND": "!app-misc/blocker-update-order-a",
            },
            "app-misc/installed-old-version-blocks-hard-a-1": {
                "EAPI": "2",
                "RDEPEND": "!!app-misc/blocker-update-order-hard-a",
            },
            "app-misc/installed-old-version-blocks-hard-unsolvable-a-1": {
                "EAPI": "2",
                "RDEPEND": "!!app-misc/blocker-update-order-hard-unsolvable-a",
            },
            "sys-apps/portage-2.1.9.42": {
                "DEPEND": "dev-lang/python",
                "RDEPEND": "dev-lang/python",
            },
            "dev-lang/python-3.1": {},
            "virtual/libc-0": {
                "RDEPEND": "sys-libs/glibc",
            },
            "sys-devel/binutils-2.18": {},
            "sys-libs/glibc-2.11": {
                "DEPEND": "virtual/os-headers sys-devel/gcc sys-devel/binutils",
                "RDEPEND": "",
            },
            "virtual/os-headers-0": {
                "RDEPEND": "sys-kernel/linux-headers",
            },
            "sys-kernel/linux-headers-2.6.38": {
                "DEPEND": "app-arch/xz-utils",
                "RDEPEND": "",
            },
            "app-arch/xz-utils-5.0.1": {},
            "media-libs/mesa-9.1.3": {
                "EAPI": "5",
                "IUSE": "+xorg",
                "USE": "xorg",
                "DEPEND": "x11-base/xorg-server:0/1.14.1=",
                "RDEPEND": "x11-base/xorg-server:0/1.14.1=",
            },
            "media-video/ffmpeg-0.7_rc1": {
                "EAPI": "2",
                "IUSE": "X +encode",
                "USE": "encode",
            },
            "virtual/ffmpeg-0.6.90": {
                "EAPI": "2",
                "IUSE": "X +encode",
                "USE": "encode",
                "RDEPEND": "|| ( >=media-video/ffmpeg-0.6.90_rc0-r2[X=,encode=] >=media-video/libav-0.6.90_rc[X=,encode=] )",
            },
            "x11-base/xorg-drivers-1.20-r2": {
                "EAPI": "7",
                "IUSE": "+video_cards_fbdev",
                "USE": "video_cards_fbdev",
                "PDEPEND": "x11-base/xorg-server x11-drivers/xf86-video-fbdev",
            },
            "x11-base/xorg-server-1.14.1": {
                "EAPI": "5",
                "SLOT": "0/1.14.1",
                "DEPEND": "media-libs/mesa",
                "RDEPEND": "media-libs/mesa",
                "PDEPEND": "x11-base/xorg-drivers",
            },
            "x11-drivers/xf86-video-fbdev-0.5.0-r1": {
                "EAPI": "7",
                "DEPEND": "x11-base/xorg-server",
                "RDEPEND": "x11-base/xorg-server:0/1.14.1=",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/circ-direct-a", "app-misc/circ-direct-b"],
                success=True,
                all_permutations=True,
                mergelist=["app-misc/circ-direct-a-1", "app-misc/circ-direct-b-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/some-app-a"],
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    (
                        "app-misc/circ-runtime-a-1",
                        "app-misc/circ-runtime-b-1",
                        "app-misc/circ-runtime-c-1",
                    ),
                    "app-misc/some-app-a-1",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/some-app-a"],
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    (
                        "app-misc/circ-runtime-c-1",
                        "app-misc/circ-runtime-b-1",
                        "app-misc/circ-runtime-a-1",
                    ),
                    "app-misc/some-app-a-1",
                ],
            ),
            # Test unsolvable circular dep that is RDEPEND in one
            # direction and DEPEND in the other.
            ResolverPlaygroundTestCase(
                ["app-misc/circ-buildtime-unsolvable-a"],
                success=False,
                circular_dependency_solutions={},
            ),
            # Test optimal merge order for a circular dep that is
            # RDEPEND in one direction and DEPEND in the other.
            # This requires an installed instance of the DEPEND
            # package in order to be solvable.
            ResolverPlaygroundTestCase(
                ["app-misc/some-app-c", "app-misc/circ-buildtime-a"],
                success=True,
                ambiguous_merge_order=True,
                # The following merge order assertion reflects optimal order for
                # a circular relationship which is DEPEND in one direction and
                # RDEPEND in the other. However, it is not respected because
                # it would result in a temporarily broken RDEPEND, so we instead
                # rely on satisfied installed build-time dependencies.
                merge_order_assertions=(
                    ("app-misc/circ-buildtime-a-1", "app-misc/circ-buildtime-c-1"),
                ),
                mergelist=[
                    (
                        "app-misc/circ-buildtime-b-1",
                        "app-misc/circ-buildtime-c-1",
                        "app-misc/circ-buildtime-a-1",
                    ),
                    "app-misc/some-app-c-1",
                ],
            ),
            # Test optimal merge order for a circular dep that is
            # RDEPEND in one direction and PDEPEND in the other.
            ResolverPlaygroundTestCase(
                ["app-misc/some-app-b"],
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    "app-misc/circ-post-runtime-a-1",
                    (
                        "app-misc/circ-post-runtime-b-1",
                        "app-misc/circ-post-runtime-c-1",
                    ),
                    "app-misc/some-app-b-1",
                ],
            ),
            # Test optimal merge order for a circular dep that is
            # RDEPEND in one direction and DEPEND in the other,
            # with all dependencies initially satisfied. Optimally,
            # the DEPEND/buildtime dep should be updated before the
            # package that depends on it, even though it's feasible
            # to update it later since it is already satisfied.
            ResolverPlaygroundTestCase(
                [
                    "app-misc/circ-satisfied-a",
                    "app-misc/circ-satisfied-b",
                    "app-misc/circ-satisfied-c",
                ],
                success=True,
                all_permutations=True,
                ambiguous_merge_order=True,
                merge_order_assertions=(
                    ("app-misc/circ-satisfied-a-1", "app-misc/circ-satisfied-c-1"),
                ),
                mergelist=[
                    (
                        "app-misc/circ-satisfied-a-1",
                        "app-misc/circ-satisfied-b-1",
                        "app-misc/circ-satisfied-c-1",
                    )
                ],
            ),
            # In the case of multiple runtime cycles, where some cycles
            # may depend on smaller independent cycles, it's optimal
            # to merge smaller independent cycles before other cycles
            # that depend on them.
            ResolverPlaygroundTestCase(
                [
                    "app-misc/circ-smallest-a",
                    "app-misc/circ-smallest-c",
                    "app-misc/circ-smallest-f",
                ],
                success=True,
                ambiguous_merge_order=True,
                all_permutations=True,
                mergelist=[
                    ("app-misc/circ-smallest-a-1", "app-misc/circ-smallest-b-1"),
                    (
                        "app-misc/circ-smallest-c-1",
                        "app-misc/circ-smallest-d-1",
                        "app-misc/circ-smallest-e-1",
                    ),
                    ("app-misc/circ-smallest-f-1", "app-misc/circ-smallest-g-1"),
                ],
            ),
            # installed package has buildtime-only blocker
            # that should be ignored
            ResolverPlaygroundTestCase(
                ["app-misc/blocker-buildtime-a"],
                success=True,
                mergelist=["app-misc/blocker-buildtime-a-1"],
            ),
            # We're installing a package that an old version of
            # an installed package blocks. However, an update is
            # available to the old package. The old package should
            # be updated first, in order to solve the blocker without
            # any need for blocking packages to temporarily overlap.
            ResolverPlaygroundTestCase(
                [
                    "app-misc/blocker-update-order-a",
                    "app-misc/installed-old-version-blocks-a",
                ],
                success=True,
                all_permutations=True,
                mergelist=[
                    "app-misc/installed-old-version-blocks-a-2",
                    "app-misc/blocker-update-order-a-1",
                ],
            ),
            # This is the same as above but with a hard blocker. The hard
            # blocker is solved automatically since the update makes it
            # irrelevant.
            ResolverPlaygroundTestCase(
                [
                    "app-misc/blocker-update-order-hard-a",
                    "app-misc/installed-old-version-blocks-hard-a",
                ],
                success=True,
                all_permutations=True,
                mergelist=[
                    "app-misc/installed-old-version-blocks-hard-a-2",
                    "app-misc/blocker-update-order-hard-a-1",
                ],
            ),
            # This is similar to the above case except that it's unsolvable
            # due to merge order, unless bug 250286 is implemented so that
            # the installed blocker will be unmerged before installation
            # of the package it blocks (rather than after like a soft blocker
            # would be handled). The "unmerge before" behavior requested
            # in bug 250286 must be optional since essential programs or
            # libraries may be temporarily unavailable during a
            # non-overlapping update like this.
            ResolverPlaygroundTestCase(
                [
                    "app-misc/blocker-update-order-hard-unsolvable-a",
                    "app-misc/installed-old-version-blocks-hard-unsolvable-a",
                ],
                success=False,
                all_permutations=True,
                ambiguous_merge_order=True,
                merge_order_assertions=(
                    (
                        "app-misc/blocker-update-order-hard-unsolvable-a-1",
                        "app-misc/installed-old-version-blocks-hard-unsolvable-a-2",
                    ),
                ),
                mergelist=[
                    (
                        "app-misc/blocker-update-order-hard-unsolvable-a-1",
                        "app-misc/installed-old-version-blocks-hard-unsolvable-a-2",
                        "!!app-misc/blocker-update-order-hard-unsolvable-a",
                    )
                ],
            ),
            # The installed package has runtime blockers that
            # should cause it to be uninstalled. The uninstall
            # task is executed only after blocking packages have
            # been merged.
            # TODO: distinguish between install/uninstall tasks in mergelist
            ResolverPlaygroundTestCase(
                ["app-misc/blocker-runtime-a", "app-misc/blocker-runtime-b"],
                success=True,
                all_permutations=True,
                ambiguous_merge_order=True,
                mergelist=[
                    ("app-misc/blocker-runtime-a-1", "app-misc/blocker-runtime-b-1"),
                    "[uninstall]app-misc/installed-blocker-a-1",
                    ("!app-misc/blocker-runtime-a", "!app-misc/blocker-runtime-b"),
                ],
            ),
            # We have a soft buildtime blocker against an installed
            # package that should cause it to be uninstalled. Note that with
            # soft blockers, the blocking packages are allowed to temporarily
            # overlap. This allows any essential programs/libraries provided
            # by both packages to be available at all times.
            # TODO: distinguish between install/uninstall tasks in mergelist
            ResolverPlaygroundTestCase(
                ["app-misc/blocker-buildtime-unbuilt-a"],
                success=True,
                mergelist=[
                    "app-misc/blocker-buildtime-unbuilt-a-1",
                    "[uninstall]app-misc/installed-blocker-a-1",
                    "!app-misc/installed-blocker-a",
                ],
            ),
            # We have a hard buildtime blocker against an installed
            # package that will not resolve automatically (unless
            # the option requested in bug 250286 is implemented).
            ResolverPlaygroundTestCase(
                ["app-misc/blocker-buildtime-unbuilt-hard-a"],
                success=False,
                mergelist=[
                    "app-misc/blocker-buildtime-unbuilt-hard-a-1",
                    "!!app-misc/installed-blocker-a",
                ],
            ),
            # An installed package has a hard runtime blocker that
            # will not resolve automatically (unless the option
            # requested in bug 250286 is implemented).
            ResolverPlaygroundTestCase(
                ["app-misc/blocker-runtime-hard-a"],
                success=False,
                mergelist=[
                    "app-misc/blocker-runtime-hard-a-1",
                    "!!app-misc/blocker-runtime-hard-a",
                ],
            ),
            # Test swapping of providers for a new-style virtual package,
            # which relies on delayed evaluation of disjunctive (virtual
            # and ||) deps as required to solve bug #264434. Note that
            # this behavior is not supported for old-style PROVIDE virtuals,
            # as reported in bug #339164.
            ResolverPlaygroundTestCase(
                ["media-video/libav"],
                success=True,
                mergelist=[
                    "media-video/libav-0.7_pre20110327",
                    "[uninstall]media-video/ffmpeg-0.7_rc1",
                    "!media-video/ffmpeg",
                ],
            ),
            # Test that OS_HEADERS_PACKAGE_ATOM and LIBC_PACKAGE_ATOM
            # are merged asap, in order to account for implicit
            # dependencies. See bug #303567. Optimally, satisfied deps
            # are always merged after the asap nodes that depend on them.
            ResolverPlaygroundTestCase(
                [
                    "app-arch/xz-utils",
                    "sys-kernel/linux-headers",
                    "sys-devel/binutils",
                    "sys-libs/glibc",
                ],
                options={"--complete-graph": True},
                success=True,
                all_permutations=True,
                ambiguous_merge_order=True,
                mergelist=[
                    "sys-kernel/linux-headers-2.6.39",
                    "sys-devel/gcc-4.5.2",
                    "sys-libs/glibc-2.13",
                    ("app-arch/xz-utils-5.0.2", "sys-devel/binutils-2.20.1"),
                ],
            ),
            # Test asap install of PDEPEND for bug #180045.
            ResolverPlaygroundTestCase(
                [
                    "kde-base/kmines",
                    "kde-base/kdnssd",
                    "kde-base/kdelibs",
                    "app-arch/xz-utils",
                ],
                success=True,
                all_permutations=True,
                ambiguous_merge_order=True,
                merge_order_assertions=(
                    ("dev-util/pkgconfig-0.25-r2", "kde-misc/kdnssd-avahi-0.1.2"),
                    ("kde-misc/kdnssd-avahi-0.1.2", "kde-base/libkdegames-3.5.7"),
                    ("kde-misc/kdnssd-avahi-0.1.2", "kde-base/kdnssd-3.5.7"),
                    ("kde-base/libkdegames-3.5.7", "kde-base/kmines-3.5.7"),
                ),
                mergelist=[
                    (
                        "kde-base/kdelibs-3.5.7",
                        "dev-util/pkgconfig-0.25-r2",
                        "kde-misc/kdnssd-avahi-0.1.2",
                        "app-arch/xz-utils-5.0.2",
                        "kde-base/libkdegames-3.5.7",
                        "kde-base/kdnssd-3.5.7",
                        "kde-base/kmines-3.5.7",
                    )
                ],
            ),
            # Test satisfied circular DEPEND/RDEPEND with one := operator.
            # Both deps are already satisfied by installed packages, but
            # the := dep is given higher priority in merge order.
            ResolverPlaygroundTestCase(
                [
                    "media-libs/mesa",
                    "x11-drivers/xf86-video-fbdev",
                    "x11-base/xorg-server",
                ],
                success=True,
                all_permutations=True,
                mergelist=[
                    "x11-base/xorg-server-1.14.1",
                    "media-libs/mesa-9.1.3",
                    "x11-drivers/xf86-video-fbdev-0.5.0-r1",
                ],
            ),
            # Test prioritization of the find_smallest_cycle function, which should
            # minimize the use of installed packages to break cycles. If installed
            # packages must be used to break cycles, then it should prefer to do this
            # for runtime dependencies over buildtime dependencies. If a package needs
            # to be uninstalled in order to solve a blocker, then it should prefer to
            # do this before it uses an installed package to break a cycle.
            ResolverPlaygroundTestCase(
                [
                    "app-misc/some-app-a",
                    "app-misc/some-app-b",
                    "app-misc/some-app-c",
                    "app-misc/circ-buildtime-a",
                    "app-misc/blocker-buildtime-unbuilt-a",
                    "media-libs/mesa",
                    "x11-base/xorg-server",
                    "app-misc/circ-direct-a",
                    "app-misc/circ-direct-b",
                    "app-misc/circ-satisfied-a",
                    "app-misc/circ-satisfied-b",
                    "app-misc/circ-satisfied-c",
                ],
                success=True,
                mergelist=[
                    "app-misc/circ-post-runtime-a-1",
                    "app-misc/circ-post-runtime-c-1",
                    "app-misc/circ-post-runtime-b-1",
                    "app-misc/some-app-b-1",
                    "app-misc/circ-runtime-a-1",
                    "app-misc/circ-runtime-c-1",
                    "app-misc/circ-runtime-b-1",
                    "app-misc/some-app-a-1",
                    "app-misc/blocker-buildtime-unbuilt-a-1",
                    "[uninstall]app-misc/installed-blocker-a-1",
                    "!app-misc/installed-blocker-a",
                    "app-misc/circ-direct-a-1",
                    "app-misc/circ-direct-b-1",
                    "x11-base/xorg-server-1.14.1",
                    "media-libs/mesa-9.1.3",
                    "app-misc/circ-buildtime-a-1",
                    "app-misc/circ-buildtime-c-1",
                    "app-misc/circ-buildtime-b-1",
                    "app-misc/some-app-c-1",
                    "app-misc/circ-satisfied-a-1",
                    "app-misc/circ-satisfied-c-1",
                    "app-misc/circ-satisfied-b-1",
                ],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.DepPriority import DepPriority


class DepPriorityNormalRange:
    """
    DepPriority properties              Index      Category

    buildtime                                      HARD
    runtime                                3       MEDIUM
    runtime_post                           2       MEDIUM_SOFT
    optional                               1       SOFT
    (none of the above)                    0       NONE
    """

    MEDIUM = 3
    MEDIUM_SOFT = 2
    MEDIUM_POST = 2
    SOFT = 1
    NONE = 0

    @classmethod
    def _ignore_optional(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        return bool(priority.optional)

    @classmethod
    def _ignore_runtime_post(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        return bool(priority.optional or priority.runtime_post)

    @classmethod
    def _ignore_runtime(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        # If we ever allow "optional" runtime_slot_op, we'll need
        # to adjust this appropriately. But only build time dependencies
        # are optional right now, so it's not an issue as-is.
        return bool(
            not (priority.runtime_slot_op and not priority.cross)
            and (priority.optional or not priority.buildtime)
        )

    ignore_medium = _ignore_runtime
    ignore_medium_soft = _ignore_runtime_post
    ignore_medium_post = _ignore_runtime_post
    ignore_soft = _ignore_optional


DepPriorityNormalRange.ignore_priority = (
    None,
    DepPriorityNormalRange._ignore_optional,
    DepPriorityNormalRange._ignore_runtime_post,
    DepPriorityNormalRange._ignore_runtime,
)



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["getmaskingstatus"]

import portage
from portage import eapi_is_supported, _eapi_is_deprecated
from portage.exception import InvalidDependString
from portage.localization import _
from portage.package.ebuild.config import config
from portage.versions import _pkg_str


class _UnmaskHint:
    __slots__ = ("key", "value")

    def __init__(self, key, value):
        self.key = key
        self.value = value


class _MaskReason:
    __slots__ = ("category", "message", "unmask_hint")

    def __init__(self, category, message, unmask_hint=None):
        self.category = category
        self.message = message
        self.unmask_hint = unmask_hint


def getmaskingstatus(mycpv, settings=None, portdb=None, myrepo=None):
    if settings is None:
        settings = config(clone=portage.settings)
    if portdb is None:
        portdb = portage.portdb

    return [
        mreason.message
        for mreason in _getmaskingstatus(mycpv, settings, portdb, myrepo)
    ]


def _getmaskingstatus(mycpv, settings, portdb, myrepo=None):
    metadata = None
    installed = False
    if not isinstance(mycpv, str):
        # emerge passed in a Package instance
        pkg = mycpv
        mycpv = pkg.cpv
        metadata = pkg._metadata
        installed = pkg.installed

    if metadata is None:
        db_keys = list(portdb._aux_cache_keys)
        try:
            metadata = dict(zip(db_keys, portdb.aux_get(mycpv, db_keys, myrepo=myrepo)))
        except KeyError:
            if not portdb.cpv_exists(mycpv):
                raise
            return [_MaskReason("corruption", "corruption")]
        if "?" in metadata["LICENSE"]:
            settings.setcpv(mycpv, mydb=metadata)
            metadata["USE"] = settings["PORTAGE_USE"]
        else:
            metadata["USE"] = ""

    try:
        mycpv.slot
    except AttributeError:
        try:
            mycpv = _pkg_str(mycpv, metadata=metadata, settings=settings)
        except portage.exception.InvalidData:
            raise ValueError(_("invalid CPV: %s") % mycpv)

    rValue = []

    # package.mask checking
    if settings._getMaskAtom(mycpv, metadata):
        rValue.append(
            _MaskReason("package.mask", "package.mask", _UnmaskHint("p_mask", None))
        )

    # keywords checking
    eapi = metadata["EAPI"]
    mygroups = settings._getKeywords(mycpv, metadata)
    licenses = metadata["LICENSE"]
    properties = metadata["PROPERTIES"]
    restrict = metadata["RESTRICT"]
    if not eapi_is_supported(eapi):
        return [_MaskReason("EAPI", f"EAPI {eapi}")]
    if _eapi_is_deprecated(eapi) and not installed:
        return [_MaskReason("EAPI", f"EAPI {eapi}")]
    egroups = settings.configdict["backupenv"].get("ACCEPT_KEYWORDS", "").split()
    global_accept_keywords = settings.get("ACCEPT_KEYWORDS", "")
    pgroups = global_accept_keywords.split()
    myarch = settings["ARCH"]
    if pgroups and myarch not in pgroups:
        """For operating systems other than Linux, ARCH is not necessarily a
        valid keyword."""
        myarch = pgroups[0].lstrip("~")

    # NOTE: This logic is copied from KeywordsManager.getMissingKeywords().
    unmaskgroups = settings._keywords_manager.getPKeywords(
        mycpv, metadata["SLOT"], metadata["repository"], global_accept_keywords
    )
    pgroups.extend(unmaskgroups)
    if unmaskgroups or egroups:
        pgroups = settings._keywords_manager._getEgroups(egroups, pgroups)
    else:
        pgroups = set(pgroups)

    kmask = "missing"
    kmask_hint = None

    if "**" in pgroups:
        kmask = None
    else:
        for keyword in pgroups:
            if keyword in mygroups:
                kmask = None
                break

    if kmask:
        for gp in mygroups:
            if gp == "*":
                kmask = None
                break
            elif gp == "~*":
                for x in pgroups:
                    if x[:1] == "~":
                        kmask = None
                        break
                if kmask is None:
                    break
            elif gp == "-" + myarch and myarch in pgroups:
                kmask = "-" + myarch
                break
            elif gp == "~" + myarch and myarch in pgroups:
                kmask = "~" + myarch
                kmask_hint = _UnmaskHint("unstable keyword", kmask)
                break

    if kmask == "missing":
        kmask_hint = _UnmaskHint("unstable keyword", "**")

    try:
        missing_licenses = settings._getMissingLicenses(mycpv, metadata)
        if missing_licenses:
            allowed_tokens = {"||", "(", ")"}
            allowed_tokens.update(missing_licenses)
            license_split = licenses.split()
            license_split = [x for x in license_split if x in allowed_tokens]
            msg = license_split[:]
            msg.append("license(s)")
            rValue.append(
                _MaskReason(
                    "LICENSE",
                    " ".join(msg),
                    _UnmaskHint("license", set(missing_licenses)),
                )
            )
    except portage.exception.InvalidDependString as e:
        rValue.append(_MaskReason("invalid", "LICENSE: " + str(e)))

    try:
        missing_properties = settings._getMissingProperties(mycpv, metadata)
        if missing_properties:
            allowed_tokens = {"||", "(", ")"}
            allowed_tokens.update(missing_properties)
            properties_split = properties.split()
            properties_split = [x for x in properties_split if x in allowed_tokens]
            msg = properties_split[:]
            msg.append("properties")
            rValue.append(_MaskReason("PROPERTIES", " ".join(msg)))
    except portage.exception.InvalidDependString as e:
        rValue.append(_MaskReason("invalid", "PROPERTIES: " + str(e)))

    try:
        missing_restricts = settings._getMissingRestrict(mycpv, metadata)
        if missing_restricts:
            msg = list(missing_restricts)
            msg.append("in RESTRICT")
            rValue.append(_MaskReason("RESTRICT", " ".join(msg)))
    except InvalidDependString as e:
        rValue.append(_MaskReason("invalid", f"RESTRICT: {e}"))

    # Only show KEYWORDS masks for installed packages
    # if they're not masked for any other reason.
    if kmask and (not installed or not rValue):
        rValue.append(
            _MaskReason("KEYWORDS", kmask + " keyword", unmask_hint=kmask_hint)
        )

    return rValue



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class AutounmaskKeepKeywordsTestCase(TestCase):
    def testAutounmaskKeepKeywordsTestCase(self):
        ebuilds = {
            "app-misc/A-2": {
                "EAPI": "6",
                "RDEPEND": "app-misc/B",
            },
            "app-misc/A-1": {
                "EAPI": "6",
                "RDEPEND": "app-misc/C[foo]",
            },
            "app-misc/B-1": {
                "EAPI": "6",
                "KEYWORDS": "~x86",
            },
            "app-misc/C-1": {
                "EAPI": "6",
                "IUSE": "foo",
            },
        }
        installed = {}

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                success=False,
                options={
                    "--autounmask-keep-keywords": "n",
                },
                mergelist=[
                    "app-misc/B-1",
                    "app-misc/A-2",
                ],
                unstable_keywords={"app-misc/B-1"},
            ),
            # --autounmask-keep-keywords prefers app-misc/A-1 because
            # it can be installed without accepting unstable
            # keywords
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                success=False,
                options={
                    "--autounmask-keep-keywords": "y",
                },
                mergelist=[
                    "app-misc/C-1",
                    "app-misc/A-1",
                ],
                use_changes={"app-misc/C-1": {"foo": True}},
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, debug=False)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2005-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

doc = """Scan for failed merges and fix them."""
__doc__ = doc


module_spec = {
    "name": "merges",
    "description": doc,
    "provides": {
        "merges": {
            "name": "merges",
            "sourcefile": "merges",
            "class": "MergesHandler",
            "description": doc,
            "functions": ["check", "fix", "purge"],
            "func_desc": {
                "purge": {
                    "short": "-P",
                    "long": "--purge",
                    "help": "Removes the list of previously failed merges."
                    + " WARNING: Only use this option if you plan on"
                    + " manually fixing them or do not want them"
                    " re-installed.",
                    "status": "Removing %s",
                    "action": "store_true",
                    "func": "purge",
                }
            },
            "opt_desc": {
                "yes": {
                    "short": "-y",
                    "long": "--yes",
                    "help": (
                        "(merges submodule only): Do not prompt for "
                        "emerge invocations"
                    ),
                    "action": "store_true",
                    "dest": "yes",
                }
            },
        }
    },
}



# Copyright 2005-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """Check and clean old logs in the PORTAGE_LOGDIR."""
__doc__ = doc


module_spec = {
    "name": "logs",
    "description": doc,
    "provides": {
        "module1": {
            "name": "logs",
            "sourcefile": "logs",
            "class": "CleanLogs",
            "description": doc,
            "functions": ["check", "clean"],
            "func_desc": {
                "clean": {
                    "short": "-C",
                    "long": "--clean",
                    "help": "Cleans out logs more than 7 days old (cleanlogs only)"
                    + "   module-options: -t, -p",
                    "status": "Cleaning %s",
                    "action": "store_true",
                    "func": "clean",
                },
                "time": {
                    "short": "-t",
                    "long": "--time",
                    "help": "(cleanlogs only): -t, --time   Delete logs older than NUM of days",
                    "status": "",
                    "type": int,
                    "dest": "NUM",
                    "func": "clean",
                },
                "pretend": {
                    "short": "-p",
                    "long": "--pretend",
                    "help": "(cleanlogs only): -p, --pretend   Output logs that would be deleted",
                    "status": "",
                    "action": "store_true",
                    "dest": "pretend",
                    "func": "clean",
                },
            },
        }
    },
}



# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import os
import stat
import tempfile

from portage.tests import TestCase
from portage.util import atomic_ofstream


class AtomicOFStreamTestCase(TestCase):
    def test_enospc_rollback(self):
        file_name = "foo"
        start_dir = os.getcwd()
        with tempfile.TemporaryDirectory() as tempdir:
            try:
                os.chdir(tempdir)
                with self.assertRaises(OSError):
                    with atomic_ofstream(file_name) as f:
                        f.write("hello")
                        raise OSError(errno.ENOSPC, "No space left on device")
                self.assertFalse(os.path.exists(file_name))
                self.assertEqual(os.listdir(tempdir), [])
            finally:
                os.chdir(start_dir)

    def test_open_failure(self):
        file_name = "bad/path"
        start_dir = os.getcwd()
        with tempfile.TemporaryDirectory() as tempdir:
            try:
                os.chdir(tempdir)
                with self.assertRaises(OSError):
                    with atomic_ofstream(file_name):
                        pass
                self.assertEqual(os.listdir(tempdir), [])
            finally:
                os.chdir(start_dir)

    def test_broken_symlink(self):
        content = "foo"
        broken_symlink = "symlink"
        symlink_targets = (("foo/bar/baz", False), ("baz", True))
        start_dir = os.getcwd()
        for symlink_target, can_follow in symlink_targets:
            with tempfile.TemporaryDirectory() as tempdir:
                try:
                    os.chdir(tempdir)
                    with open(broken_symlink, "w") as f:
                        default_file_mode = stat.S_IMODE(os.fstat(f.fileno()).st_mode)
                        os.unlink(broken_symlink)
                    os.symlink(symlink_target, broken_symlink)
                    with atomic_ofstream(broken_symlink) as f:
                        f.write(content)
                    with open(broken_symlink) as f:
                        self.assertEqual(f.read(), content)
                    self.assertEqual(os.path.islink(broken_symlink), can_follow)
                    self.assertEqual(
                        stat.S_IMODE(os.stat(broken_symlink).st_mode), default_file_mode
                    )
                finally:
                    os.chdir(start_dir)

    def test_preserves_mode(self):
        file_name = "foo"
        file_mode = 0o604
        start_dir = os.getcwd()
        with tempfile.TemporaryDirectory() as tempdir:
            try:
                os.chdir(tempdir)
                with open(file_name, "wb"):
                    pass
                self.assertNotEqual(stat.S_IMODE(os.stat(file_name).st_mode), file_mode)
                os.chmod(file_name, file_mode)
                st_before = os.stat(file_name)
                self.assertEqual(stat.S_IMODE(st_before.st_mode), file_mode)
                with atomic_ofstream(file_name):
                    pass
                st_after = os.stat(file_name)
                self.assertNotEqual(st_before.st_ino, st_after.st_ino)
                self.assertEqual(stat.S_IMODE(st_after.st_mode), file_mode)
            finally:
                os.chdir(start_dir)



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import pickle

from portage.tests import TestCase
from _emerge.Package import _PackageMetadataWrapperBase
from _emerge.FifoIpcDaemon import FifoIpcDaemon


class PickleTestCase(TestCase):
    def test_PackageMetadataWrapperBase(self):
        """
        Verify that instances of slot_dict_class, like
        PackageMetadataWrapperBase, are picklable for
        compatibility with the multiprocessing spawn
        start method.
        """
        obj = _PackageMetadataWrapperBase(EAPI="8")
        self.assertEqual(obj["EAPI"], "8")
        serialized = pickle.dumps(obj)
        obj_copy = pickle.loads(serialized)
        self.assertEqual(len(obj_copy), len(obj))
        self.assertEqual(obj_copy["EAPI"], obj["EAPI"])
        self.assertEqual(obj_copy, obj)

    def test_FifoIpcDaemon_files_dict(self):
        """
        Verify that FifoIpcDaemon._files_dict instances are picklable for
        compatibility with the multiprocessing spawn start method.
        """
        obj = FifoIpcDaemon._files_dict(
            (k, "test-value") for k in FifoIpcDaemon._file_names
        )
        self.assertEqual(obj["pipe_in"], "test-value")
        # Attributes of same name exist because of slot_dict_class prefix="" argument.
        self.assertEqual(obj.pipe_in, obj["pipe_in"])
        serialized = pickle.dumps(obj)
        obj_copy = pickle.loads(serialized)
        self.assertEqual(len(obj_copy), len(obj))
        self.assertEqual(obj_copy["pipe_in"], obj["pipe_in"])
        self.assertEqual(obj_copy.pipe_in, obj["pipe_in"])
        self.assertEqual(obj_copy, obj)
